from datetime import datetime
from enum import Enum
from typing import Generic, Protocol, TypeVar, Union

from pydantic import BaseModel, Field, JsonValue, ValidationInfo, field_validator

from exponent.core.types.command_data import (
    DEFAULT_CODE_BLOCK_TIMEOUT,
    WRITE_STRATEGY_NATURAL_EDIT,
    CommandDataType,
    EditContent,
    FileWriteStrategyName,
    NaturalEditContent,
)


class FileWriteErrorType(str, Enum):
    TERMINATION_REQUESTED = "TERMINATION_REQUESTED"
    NO_OP = "NO_OP"
    FAILED_APPLY = "FAILED_APPLY"
    FAILED_GENERATION = "FAILED_GENERATION"
    CLI_DISCONNECTED = "CLI_DISCONNECTED"


class ExponentEvent(BaseModel):
    chat_uuid: str
    event_uuid: str
    parent_uuid: str | None
    turn_uuid: str

    metadata: dict[str, JsonValue] = Field(default_factory=dict)


class PersistedExponentEvent(ExponentEvent):
    db_timestamp: datetime | None = None


class CodeBlockEvent(PersistedExponentEvent):
    language: str
    content: str
    timeout: int = DEFAULT_CODE_BLOCK_TIMEOUT
    require_confirmation: bool = False


class FileWriteEvent(PersistedExponentEvent):
    file_path: str
    language: str
    write_strategy: FileWriteStrategyName
    write_content: NaturalEditContent | EditContent
    content: str
    error_content: str | None
    error_type: FileWriteErrorType | None
    require_confirmation: bool = False

    @field_validator("write_content")
    def validate_write_content_type(
        cls, v: NaturalEditContent | EditContent, info: ValidationInfo
    ) -> NaturalEditContent | EditContent:
        write_strategy = info.data.get("write_strategy")
        if write_strategy == WRITE_STRATEGY_NATURAL_EDIT:
            if not isinstance(v, NaturalEditContent):
                raise ValueError(
                    "When write_strategy is NATURAL_EDIT, write_content must be NaturalEditContent"
                )
        elif not isinstance(v, EditContent):
            raise ValueError(
                "For non-NATURAL_EDIT strategies, write_content must be EditContent"
            )
        return v


T = TypeVar("T", bound=CommandDataType)


class HoldsCommandData(Protocol, Generic[T]):
    data: T


class CommandEvent(PersistedExponentEvent):
    data: CommandDataType
    require_confirmation: bool = False


class MultiCommandEvent(PersistedExponentEvent):
    data: list[CommandDataType]
    require_confirmation: bool = False


LocalEventType = Union[FileWriteEvent, CodeBlockEvent, CommandEvent]



import msgspec
import yaml


class PartialToolResult(msgspec.Struct, tag_field="tool_name", omit_defaults=True):
    pass


class ToolInput(msgspec.Struct, tag_field="tool_name", omit_defaults=True):
    """Concrete subclasses describe the full input schema for a tool."""


class ToolResult(msgspec.Struct, tag_field="tool_name", omit_defaults=True):
    """Concrete subclasses return data from a tool execution."""

    def to_text(self) -> str:
        """
        This provides a default textual representation of the tool result. Override it as needed for your tool."""
        d = msgspec.to_builtins(self)
        del d["tool_name"]
        return yaml.dump(d)


class ErrorToolResult(ToolResult, tag="error"):
    error_message: str


READ_TOOL_NAME = "read"


class ReadToolInput(ToolInput, tag=READ_TOOL_NAME):
    file_path: str
    offset: int | None = None
    limit: int | None = None


class ReadToolResult(ToolResult, tag=READ_TOOL_NAME):
    content: str
    num_lines: int
    start_line: int
    total_lines: int

    def to_text(self) -> str:
        lines = self.content.splitlines()
        lines = [
            f"{str(i).rjust(6)}→{line}"
            for i, line in enumerate(lines, start=self.start_line + 1)
        ]
        return "\n".join(lines)


LIST_TOOL_NAME = "ls"


class ListToolInput(ToolInput, tag=LIST_TOOL_NAME):
    path: str
    ignore: list[str] | None = None


class ListToolResult(ToolResult, tag=LIST_TOOL_NAME):
    files: list[str]
    truncated: bool = False


GLOB_TOOL_NAME = "glob"


class GlobToolInput(ToolInput, tag=GLOB_TOOL_NAME):
    pattern: str
    path: str | None = None


class GlobToolResult(ToolResult, tag=GLOB_TOOL_NAME):
    filenames: list[str]
    duration_ms: int
    num_files: int
    truncated: bool


WRITE_TOOL_NAME = "write"


class WriteToolInput(ToolInput, tag=WRITE_TOOL_NAME):
    file_path: str
    content: str


class WriteToolResult(ToolResult, tag=WRITE_TOOL_NAME):
    message: str


GREP_TOOL_NAME = "grep"


class GrepToolInput(ToolInput, tag=GREP_TOOL_NAME):
    pattern: str
    path: str | None = None
    include: str | None = None


class GrepToolResult(ToolResult, tag=GREP_TOOL_NAME):
    matches: list[str]
    truncated: bool = False


BASH_TOOL_NAME = "bash"


class BashToolInput(ToolInput, tag=BASH_TOOL_NAME):
    command: str
    timeout: int | None = None
    description: str | None = None


class PartialBashToolResult(PartialToolResult, tag=BASH_TOOL_NAME):
    shell_output: str | None = None


class BashToolResult(ToolResult, tag=BASH_TOOL_NAME):
    shell_output: str
    duration_ms: int
    exit_code: int | None
    timed_out: bool
    stopped_by_user: bool


ToolInputType = (
    ReadToolInput
    | WriteToolInput
    | ListToolInput
    | GlobToolInput
    | GrepToolInput
    | BashToolInput
)
PartialToolResultType = PartialBashToolResult

ToolResultType = (
    ReadToolResult
    | WriteToolResult
    | ListToolResult
    | GlobToolResult
    | GrepToolResult
    | BashToolResult
    | ErrorToolResult
)


class ToolExecutionRequest(msgspec.Struct, tag="tool_execution"):
    tool_input: ToolInputType


class GetAllFilesRequest(msgspec.Struct, tag="get_all_files"):
    pass


class TerminateRequest(msgspec.Struct, tag="terminate"):
    pass


class GetAllFilesResponse(msgspec.Struct, tag="get_all_files"):
    files: list[str]


class TerminateResponse(msgspec.Struct, tag="terminate"):
    pass


class CliRpcRequest(msgspec.Struct):
    request_id: str
    request: ToolExecutionRequest | GetAllFilesRequest | TerminateRequest


class ToolExecutionResponse(msgspec.Struct, tag="tool_execution"):
    tool_result: ToolResultType


class ErrorResponse(msgspec.Struct, tag="error"):
    error_message: str


class CliRpcResponse(msgspec.Struct):
    request_id: str
    response: (
        ToolExecutionResponse | GetAllFilesResponse | ErrorResponse | TerminateResponse
    )



GET_CLOUD_CONFIGS_QUERY: str = """
    query {
        cloudConfigs {
            __typename
            ... on UnauthenticatedError {
                message
            }
            ... on CloudConfigs {
                configs {
                    cloudConfigUuid
                    githubOrgName
                    githubRepoName
                    setupCommands
                    repoUrl
                }
            }
        }
    }
"""

CREATE_CLOUD_CONFIG_MUTATION: str = """
    mutation CreateCloudConfig(
        $githubOrgName: String!,
        $githubRepoName: String!,
        $setupCommands: [String!],
    ) {
        createCloudConfig(
            input: {
                githubOrgName: $githubOrgName,
                githubRepoName: $githubRepoName,
                setupCommands: $setupCommands,
            }
        ) {
            __typename
            ... on UnauthenticatedError {
                message
            }
            ... on CloudConfig {
                cloudConfigUuid
                githubOrgName
                githubRepoName
                setupCommands
                repoUrl
            }
        }
    }
"""

UPDATE_CLOUD_CONFIG_MUTATION: str = """
    mutation UpdateCloudConfig(
        $cloudConfigUuid: String!,
        $githubOrgName: String!,
        $githubRepoName: String!,
        $setupCommands: [String!],
    ) {
        updateCloudConfig(
            cloudConfigUuid: $cloudConfigUuid,
            input: {
                githubOrgName: $githubOrgName,
                githubRepoName: $githubRepoName,
                setupCommands: $setupCommands,
            }
        ) {
            __typename
            ... on UnauthenticatedError {
                message
            }
            ... on CloudConfig {
                cloudConfigUuid
                githubOrgName
                githubRepoName
                setupCommands
                repoUrl
            }
        }
    }
"""



from __future__ import annotations

import enum
import json
import logging
import os
from functools import lru_cache
from importlib.metadata import Distribution, PackageNotFoundError
from typing import Any, Dict  # noqa: UP035

from pydantic import BaseModel
from pydantic_settings import (
    BaseSettings,
    JsonConfigSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)

logger = logging.getLogger(__name__)

# If the package is editable, we want to use:
# base_url = localhost:3000
# base_api_url = localhost:8000


def is_editable_install() -> bool:
    if os.getenv("ENVIRONMENT") == "test" or os.getenv("EXPONENT_TEST_AUTO_UPGRADE"):
        # We should explicitly set these variables
        # in test when needed
        return False

    try:
        dist = Distribution.from_name("indent-ai")
    except PackageNotFoundError:
        logger.info("No distribution info found for indent-ai")
        return False

    direct_url = dist.read_text("direct_url.json")
    if not direct_url:
        return False

    try:
        direct_url_json = json.loads(direct_url)
    except json.JSONDecodeError:
        logger.warning("Failed to decode distribution info for exponent-run")
        return False

    pkg_is_editable = direct_url_json.get("dir_info", {}).get("editable", False)
    return bool(pkg_is_editable)


class Environment(str, enum.Enum):
    test = "test"
    development = "development"
    staging = "staging"
    production = "production"

    @property
    def is_local(self) -> bool:
        return self in [Environment.development, Environment.test]


class APIKeys(BaseModel):
    development: str | None = None
    staging: str | None = None
    production: str | None = None


SETTINGS_FOLDER = os.path.expanduser("~/.config/exponent")
SETTINGS_FILE_PATH = os.path.join(SETTINGS_FOLDER, "config.json")


class GlobalExponentOptions(BaseModel):
    git_warning_disabled: bool = False
    auto_upgrade: bool = True
    base_api_url_override: str | None = None
    base_ws_url_override: str | None = None
    use_default_colors: bool = False


class Settings(BaseSettings):
    environment: Environment
    base_url: str
    base_api_url: str
    base_ws_url: str

    exponent_api_key: str | None = None
    extra_exponent_api_keys: Dict[str, str] = {}  # noqa: UP006

    options: GlobalExponentOptions = GlobalExponentOptions()

    log_level: str = "WARNING"

    model_config = SettingsConfigDict(
        json_file=SETTINGS_FILE_PATH,
        json_file_encoding="utf-8",
        env_prefix="EXPONENT_",
    )

    def get_base_api_url(self) -> str:
        return self.options.base_api_url_override or self.base_api_url

    def get_base_ws_url(self) -> str:
        return self.options.base_ws_url_override or self.base_ws_url

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return env_settings, JsonConfigSettingsSource(settings_cls), init_settings

    def write_settings_to_config_file(self) -> None:
        os.makedirs(SETTINGS_FOLDER, exist_ok=True)
        with open(SETTINGS_FILE_PATH, "w") as f:
            f.write(
                self.model_dump_json(
                    include=self.config_file_keys, exclude_defaults=True
                )
            )

    def get_config_file_settings(self) -> dict[str, Any]:
        return self.model_dump(include=self.config_file_keys, exclude_defaults=True)

    @property
    def config_file_path(self) -> str:
        return SETTINGS_FILE_PATH

    @property
    def config_file_keys(self) -> set[str]:
        return {"exponent_api_key", "extra_exponent_api_keys", "options"}

    @property
    def api_key(self) -> str | None:
        if self.environment.is_local:
            return self.extra_exponent_api_keys.get(Environment.development)
        elif self.environment == Environment.staging:
            return self.extra_exponent_api_keys.get(Environment.staging)
        elif self.environment == Environment.production:
            return self.exponent_api_key
        else:
            raise ValueError(f"Unknown environment: {self.environment}")

    def update_api_key(self, api_key: str) -> None:
        if self.environment == Environment.development:
            self.extra_exponent_api_keys[Environment.development] = api_key
        elif self.environment == Environment.staging:
            self.extra_exponent_api_keys[Environment.staging] = api_key
        elif self.environment == Environment.production:
            self.exponent_api_key = api_key


@lru_cache(maxsize=1)
def get_settings(use_prod: bool = False, use_staging: bool = False) -> Settings:
    if is_editable_install() and not (use_prod or use_staging):
        base_url = "http://localhost:3000"
        base_api_url = "http://localhost:8000"
        base_ws_url = "ws://localhost:8000"
        environment = Environment.development
    elif use_staging:
        base_url = "https://staging.indent.com"
        base_api_url = "https://staging-api.indent.com"
        base_ws_url = "wss://ws-staging-api.indent.com"
        environment = Environment.staging
    else:
        base_url = "https://app.indent.com"
        base_api_url = "https://api.indent.com"
        base_ws_url = "wss://ws-api.indent.com"
        environment = Environment.production

    return Settings(
        base_url=base_url,
        base_api_url=base_api_url,
        base_ws_url=base_ws_url,
        environment=environment,
    )



import asyncio

import click
from websockets.exceptions import ConnectionClosed

from exponent.commands.settings import use_settings
from exponent.commands.shell_commands import LiveView
from exponent.commands.theme import get_theme
from exponent.commands.types import exponent_cli_group
from exponent.core.config import Settings
from exponent.core.graphql.client import GraphQLClient
from exponent.core.graphql.subscriptions import INDENT_CHAT_EVENT_STREAM_SUBSCRIPTION


@exponent_cli_group(hidden=True)
def listen_cli() -> None:
    pass


@listen_cli.command()
@click.option("--chat-id", help="ID of the chat to listen to", required=True)
@click.option(
    "--known-event-uuids",
    help="Comma-separated list of known event UUIDs to skip",
    default="",
)
@use_settings
def listen(settings: Settings, chat_id: str, known_event_uuids: str) -> None:
    """Listen to events from an indent chat session."""
    api_key = settings.api_key
    if not api_key:
        raise click.UsageError("API key is not set")

    base_api_url = settings.get_base_api_url()
    base_ws_url = settings.get_base_ws_url()

    gql_client = GraphQLClient(api_key, base_api_url, base_ws_url)

    # Parse known event UUIDs
    known_uuids = []
    if known_event_uuids:
        known_uuids = [
            uuid.strip() for uuid in known_event_uuids.split(",") if uuid.strip()
        ]

    # Set up the live view with theme
    theme = get_theme(settings.options.use_default_colors)
    live_view = LiveView(theme, render_user_messages=True)

    asyncio.run(_listen(gql_client, chat_id, known_uuids, live_view))


async def _listen(
    gql_client: GraphQLClient,
    chat_id: str,
    known_event_uuids: list[str],
    live_view: LiveView,
) -> None:
    """Internal listen function that handles the WebSocket subscription."""
    while True:
        try:
            variables = {
                "chatUuid": chat_id,
                "knownFullEventUuids": known_event_uuids if known_event_uuids else None,
            }

            async for response in gql_client.subscribe(
                INDENT_CHAT_EVENT_STREAM_SUBSCRIPTION, variables
            ):
                event = response["indentChatEventStream"]
                kind = event["__typename"]

                # Handle different event types
                if kind in ["UserEvent", "AssistantEvent", "SystemEvent"]:
                    live_view.render_event(kind, event)
                elif kind == "Error":
                    print(f"Error: {event.get('message', 'Unknown error')}")
                elif kind == "UnauthenticatedError":
                    print(
                        f"Authentication error: {event.get('message', 'Unauthenticated')}"
                    )
                    break
                else:
                    print(f"Unknown event type: {kind}")

        except ConnectionClosed:
            print("WebSocket disconnected, reconnecting...")
            await asyncio.sleep(1)
        except KeyboardInterrupt:
            print("\nDisconnecting...")
            break
        except Exception as e:
            print(f"Error: {e}")
            await asyncio.sleep(1)



import json
import sys

import click

from exponent.commands.common import (
    redirect_to_login,
    refresh_api_key_task,
    run_until_complete,
    set_login_complete,
)
from exponent.commands.settings import use_settings
from exponent.commands.types import exponent_cli_group
from exponent.core.config import Settings
from exponent.core.graphql.client import GraphQLClient
from exponent.core.graphql.cloud_config_queries import (
    CREATE_CLOUD_CONFIG_MUTATION,
    GET_CLOUD_CONFIGS_QUERY,
    UPDATE_CLOUD_CONFIG_MUTATION,
)
from exponent.core.graphql.get_chats_query import GET_CHATS_QUERY
from exponent.core.graphql.github_config_queries import (
    CHECK_GITHUB_CONFIG_VALIDITY_QUERY,
    CREATE_GITHUB_CONFIG_MUTATION,
    REPOS_FOR_GITHUB_CONFIG_QUERY,
)
from exponent.core.graphql.subscriptions import AUTHENTICATED_USER_SUBSCRIPTION
from exponent.utils.version import (
    get_installed_metadata,
    get_installed_version,
    get_installer,
    get_python_path,
    get_sys_executable,
)


@exponent_cli_group()
def config_cli() -> None:
    """Manage Exponent configuration settings."""
    pass


@config_cli.command(hidden=True)
@use_settings
def debug(
    settings: Settings,
) -> None:
    click.echo("Settings:")
    click.echo(settings)
    click.echo("\nInstalled Version:")
    click.echo(get_installed_version())
    click.echo("\nPython Version:")
    click.echo(sys.version)
    click.echo("\nPython Path:")
    click.echo(get_python_path())
    click.echo("\nSys Executable Path:")
    click.echo(get_sys_executable())
    click.echo("\nInstaller:")
    click.echo(get_installer())
    click.echo("\nInstalled Metadata:")
    click.echo(get_installed_metadata())


@config_cli.command(hidden=True)
@use_settings
def check_github_config_validity(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        check_github_config_validity_task(
            api_key=settings.api_key,
            base_api_url=settings.get_base_api_url(),
            base_ws_url=settings.get_base_ws_url(),
        )
    )


async def check_github_config_validity_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    result = await graphql_client.execute(CHECK_GITHUB_CONFIG_VALIDITY_QUERY)
    click.echo(result)


@config_cli.command(hidden=True)
@use_settings
def repos_for_github_config(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        repos_for_github_config_task(
            api_key=settings.api_key,
            base_api_url=settings.get_base_api_url(),
            base_ws_url=settings.get_base_ws_url(),
        )
    )


async def repos_for_github_config_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    try:
        click.echo("Sending request to fetch repos...")
        result = await graphql_client.execute(
            REPOS_FOR_GITHUB_CONFIG_QUERY, timeout=120
        )  # 120 seconds timeout
        click.echo("Request completed. Result:")
        click.echo(result)
    except Exception as e:
        click.echo(f"An error occurred while fetching repos: {e!s}")
        click.echo(f"Error type: {type(e).__name__}")
        # Add more detailed error information if available
        if hasattr(e, "response"):
            click.echo(f"Response status: {e.response.status_code}")
            click.echo(f"Response content: {e.response.text}")


@config_cli.command(hidden=True)
@click.option(
    "--set-git-warning-disabled",
    is_flag=True,
    help="Disable the git warning for running Exponent from a non-git repository",
)
@click.option(
    "--set-git-warning-enabled",
    is_flag=True,
    help="Enable the git warning for running Exponent from a non-git repository",
)
@click.option(
    "--set-auto-upgrade-enabled",
    is_flag=True,
    help="Enable automatic upgrades",
)
@click.option(
    "--set-auto-upgrade-disabled",
    is_flag=True,
    help="Disable automatic upgrades",
)
@click.option(
    "--set-base-api-url-override",
    required=False,
    hidden=True,
    help="Override base API URL",
)
@click.option(
    "--clear-base-api-url-override",
    is_flag=True,
    hidden=True,
    help="Clear base API URL override",
)
@click.option(
    "--set-base-ws-url-override",
    required=False,
    hidden=True,
    help="Override base WS URL",
)
@click.option(
    "--clear-base-ws-url-override",
    is_flag=True,
    hidden=True,
    help="Clear base WS URL override",
)
@click.option(
    "--set-use-default-colors",
    is_flag=True,
    help="Use default colors",
)
@click.option(
    "--clear-use-default-colors",
    is_flag=True,
    help="Clear use default colors",
)
@use_settings
def config(
    settings: Settings,
    set_git_warning_disabled: bool,
    set_git_warning_enabled: bool,
    set_auto_upgrade_enabled: bool,
    set_auto_upgrade_disabled: bool,
    clear_base_api_url_override: bool,
    clear_base_ws_url_override: bool,
    set_use_default_colors: bool,
    clear_use_default_colors: bool,
    set_base_api_url_override: str | None = None,
    set_base_ws_url_override: str | None = None,
) -> None:
    """Display current Exponent configuration."""

    num_options_set = sum(
        [
            set_git_warning_disabled,
            set_git_warning_enabled,
            set_auto_upgrade_enabled,
            set_auto_upgrade_disabled,
            clear_base_ws_url_override,
            clear_base_api_url_override,
            set_base_api_url_override is not None,
            set_base_ws_url_override is not None,
            set_use_default_colors,
            clear_use_default_colors,
        ]
    )
    if num_options_set == 0:
        click.secho(
            json.dumps(settings.get_config_file_settings(), indent=2),
            fg="green",
        )
        return

    if num_options_set > 1:
        click.secho("Cannot set multiple options at the same time.", fg="red")
        return

    if set_git_warning_enabled:
        settings.options.git_warning_disabled = False
        settings.write_settings_to_config_file()

        click.secho(
            "Git warning enabled. Exponent will now check for a git repository.\n",
            fg="yellow",
        )

    if set_git_warning_disabled:
        settings.options.git_warning_disabled = True
        settings.write_settings_to_config_file()

        click.secho(
            "Git warning disabled. Exponent will no longer check for a git repository.\n",
            fg="yellow",
        )

    if set_auto_upgrade_enabled:
        settings.options.auto_upgrade = True
        settings.write_settings_to_config_file()

        click.secho(
            "Automatic upgrades enabled. Exponent will now check for updates.\n",
            fg="yellow",
        )

    if set_auto_upgrade_disabled:
        settings.options.auto_upgrade = False
        settings.write_settings_to_config_file()

        click.secho(
            "Automatic upgrades disabled. Exponent will no longer check for updates.\n",
            fg="yellow",
        )

    if clear_base_api_url_override:
        settings.options.base_api_url_override = None
        settings.write_settings_to_config_file()

        click.secho(
            "API URL override cleared.",
            fg="yellow",
        )

    if clear_base_ws_url_override:
        settings.options.base_ws_url_override = None
        settings.write_settings_to_config_file()

        click.secho(
            "WS URL override cleared.",
            fg="yellow",
        )

    if set_base_api_url_override:
        settings.options.base_api_url_override = set_base_api_url_override
        settings.write_settings_to_config_file()

        click.secho(
            "API URL override set.",
            fg="yellow",
        )

    if set_base_ws_url_override:
        settings.options.base_ws_url_override = set_base_ws_url_override
        settings.write_settings_to_config_file()

        click.secho(
            "WS URL override set.",
            fg="yellow",
        )

    if set_use_default_colors:
        settings.options.use_default_colors = True
        settings.write_settings_to_config_file()

        click.secho(
            "Use default colors set.",
            fg="yellow",
        )

    if clear_use_default_colors:
        settings.options.use_default_colors = False
        settings.write_settings_to_config_file()


@config_cli.command()
@click.option("--key", help="Your Exponent API Key")
@use_settings
def login(settings: Settings, key: str) -> None:
    """Log in to Exponent."""

    if not key:
        redirect_to_login(settings, "provided")
        return

    click.echo("Verifying API key...")
    run_until_complete(
        set_login_complete(key, settings.get_base_api_url(), settings.get_base_ws_url())
    )
    click.secho("Success!", fg="green")

    click.echo(f"Saving API Key to {settings.config_file_path}")

    if settings.api_key and settings.api_key != key:
        click.confirm("Detected existing API Key, continue? ", default=True, abort=True)

    settings.update_api_key(key)
    settings.write_settings_to_config_file()

    click.echo("API Key saved.")


@config_cli.command(hidden=True)
@use_settings
def get_chats(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        get_chats_task(
            api_key=settings.api_key,
            base_api_url=settings.get_base_api_url(),
            base_ws_url=settings.get_base_ws_url(),
        )
    )


@config_cli.command(hidden=True)
@use_settings
def get_authenticated_user(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        get_authenticated_user_task(
            api_key=settings.api_key,
            base_api_url=settings.get_base_api_url(),
            base_ws_url=settings.get_base_ws_url(),
        )
    )


async def get_chats_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    result = await graphql_client.execute(GET_CHATS_QUERY)
    click.echo(result)


async def get_authenticated_user_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    async for it in graphql_client.subscribe(AUTHENTICATED_USER_SUBSCRIPTION):
        click.echo(it)


@config_cli.command(hidden=True)
@use_settings
def get_cloud_configs(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        get_cloud_configs_task(
            api_key=settings.api_key,
            base_api_url=settings.get_base_api_url(),
            base_ws_url=settings.get_base_ws_url(),
        )
    )


async def get_cloud_configs_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    result = await graphql_client.execute(GET_CLOUD_CONFIGS_QUERY)
    click.echo(result)


async def create_github_config_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
    github_pat: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    variables = {
        "githubPat": github_pat,
    }
    result = await graphql_client.execute(CREATE_GITHUB_CONFIG_MUTATION, variables)
    click.echo(result)


@config_cli.command(hidden=True)
@click.option("--github-pat", required=True, help="Github personal access token")
@use_settings
def create_github_config(
    settings: Settings,
    github_pat: str,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        create_github_config_task(
            api_key=settings.api_key,
            base_api_url=settings.get_base_api_url(),
            base_ws_url=settings.get_base_ws_url(),
            github_pat=github_pat,
        )
    )


@config_cli.command(hidden=True)
@click.option("--github-org-name", required=True, help="GitHub organization name")
@click.option("--github-repo-name", required=True, help="GitHub repository name")
@click.option(
    "--setup_commands",
    required=False,
    help="List of commands to set up and build your repo",
    multiple=True,
)
@use_settings
def create_cloud_config(
    settings: Settings,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: list[str] | None = None,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        create_cloud_config_task(
            api_key=settings.api_key,
            base_api_url=settings.get_base_api_url(),
            base_ws_url=settings.get_base_ws_url(),
            github_org_name=github_org_name,
            github_repo_name=github_repo_name,
            setup_commands=setup_commands,
        )
    )


async def create_cloud_config_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: list[str] | None,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    variables = {
        "githubOrgName": github_org_name,
        "githubRepoName": github_repo_name,
        "setupCommands": setup_commands,
    }
    result = await graphql_client.execute(CREATE_CLOUD_CONFIG_MUTATION, variables)
    click.echo(result)


@config_cli.command(hidden=True)
@click.option("--cloud-config-uuid", required=True, help="Cloud config UUID")
@click.option("--github-org-name", required=True, help="GitHub organization name")
@click.option("--github-repo-name", required=True, help="GitHub repository name")
@click.option(
    "--setup_commands",
    required=False,
    help="List of commands to set up and build your repo",
    multiple=True,
)
@use_settings
def update_cloud_config(
    settings: Settings,
    cloud_config_uuid: str,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: list[str] | None = None,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        update_cloud_config_task(
            api_key=settings.api_key,
            base_api_url=settings.get_base_api_url(),
            base_ws_url=settings.get_base_ws_url(),
            cloud_config_uuid=cloud_config_uuid,
            github_org_name=github_org_name,
            github_repo_name=github_repo_name,
            setup_commands=setup_commands,
        )
    )


async def update_cloud_config_task(
    api_key: str,
    base_api_url: str,
    base_ws_url: str,
    cloud_config_uuid: str,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: list[str] | None,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url, base_ws_url)
    variables = {
        "cloudConfigUuid": cloud_config_uuid,
        "githubOrgName": github_org_name,
        "githubRepoName": github_repo_name,
        "setupCommands": setup_commands,
    }

    result = await graphql_client.execute(UPDATE_CLOUD_CONFIG_MUTATION, variables)
    click.echo(result)


@config_cli.command(hidden=True)
@use_settings
def refresh_key(settings: Settings) -> None:
    """Refresh your API key."""
    if not settings.api_key:
        redirect_to_login(settings)
        return

    click.echo("Refreshing API key...")
    run_until_complete(
        refresh_api_key_task(
            api_key=settings.api_key,
            base_api_url=settings.get_base_api_url(),
            base_ws_url=settings.get_base_ws_url(),
        )
    )



"""Generalized truncation framework for tool results."""

from abc import ABC, abstractmethod
from typing import TypeVar, cast

from msgspec.structs import replace

from exponent.core.remote_execution.cli_rpc_types import (
    BashToolResult,
    ErrorToolResult,
    GlobToolResult,
    GrepToolResult,
    ListToolResult,
    ReadToolResult,
    ToolResult,
    WriteToolResult,
)
from exponent.core.remote_execution.utils import truncate_output

DEFAULT_CHARACTER_LIMIT = 90_000
DEFAULT_LIST_ITEM_LIMIT = 1000
DEFAULT_LIST_PREVIEW_ITEMS = 10


class TruncationStrategy(ABC):

    @abstractmethod
    def should_truncate(self, result: ToolResult) -> bool:
        pass

    @abstractmethod
    def truncate(self, result: ToolResult) -> ToolResult:
        pass


class StringFieldTruncation(TruncationStrategy):

    def __init__(
        self,
        field_name: str,
        character_limit: int = DEFAULT_CHARACTER_LIMIT,
    ):
        self.field_name = field_name
        self.character_limit = character_limit

    def should_truncate(self, result: ToolResult) -> bool:
        if hasattr(result, self.field_name):
            value = getattr(result, self.field_name)
            if isinstance(value, str):
                return len(value) > self.character_limit
        return False

    def truncate(self, result: ToolResult) -> ToolResult:
        if not hasattr(result, self.field_name):
            return result

        value = getattr(result, self.field_name)
        if not isinstance(value, str):
            return result

        truncated_value, was_truncated = truncate_output(value, self.character_limit)

        updates = {self.field_name: truncated_value}
        if hasattr(result, "truncated") and was_truncated:
            updates["truncated"] = True

        return replace(result, **updates)


class ListFieldTruncation(TruncationStrategy):

    def __init__(
        self,
        field_name: str,
        item_limit: int = DEFAULT_LIST_ITEM_LIMIT,
        preview_items: int = DEFAULT_LIST_PREVIEW_ITEMS,
    ):
        self.field_name = field_name
        self.item_limit = item_limit
        self.preview_items = preview_items

    def should_truncate(self, result: ToolResult) -> bool:
        if hasattr(result, self.field_name):
            value = getattr(result, self.field_name)
            if isinstance(value, list):
                return len(value) > self.item_limit
        return False

    def truncate(self, result: ToolResult) -> ToolResult:
        if not hasattr(result, self.field_name):
            return result

        value = getattr(result, self.field_name)
        if not isinstance(value, list):
            return result

        total_items = len(value)
        if total_items <= self.item_limit:
            return result

        truncated_count = max(0, total_items - 2 * self.preview_items)
        truncated_list = (
            value[: self.preview_items]
            + [f"... {truncated_count} items truncated ..."]
            + value[-self.preview_items :]
        )

        updates = {self.field_name: truncated_list}
        if hasattr(result, "truncated"):
            updates["truncated"] = True

        return replace(result, **updates)


class CompositeTruncation(TruncationStrategy):

    def __init__(self, strategies: list[TruncationStrategy]):
        self.strategies = strategies

    def should_truncate(self, result: ToolResult) -> bool:
        return any(strategy.should_truncate(result) for strategy in self.strategies)

    def truncate(self, result: ToolResult) -> ToolResult:
        for strategy in self.strategies:
            if strategy.should_truncate(result):
                result = strategy.truncate(result)
        return result


TRUNCATION_REGISTRY: dict[type[ToolResult], TruncationStrategy] = {
    ReadToolResult: StringFieldTruncation("content"),
    WriteToolResult: StringFieldTruncation("message"),
    BashToolResult: StringFieldTruncation("shell_output"),
    GrepToolResult: ListFieldTruncation("matches"),
    GlobToolResult: ListFieldTruncation("filenames"),
    ListToolResult: ListFieldTruncation("files"),
}


T = TypeVar("T", bound=ToolResult)


def truncate_tool_result(result: T) -> T:
    if isinstance(result, ErrorToolResult):
        return result

    result_type = type(result)
    if result_type in TRUNCATION_REGISTRY:
        strategy = TRUNCATION_REGISTRY[result_type]
        if strategy.should_truncate(result):
            return cast(T, strategy.truncate(result))

    return result


def register_truncation_strategy(
    result_type: type[ToolResult],
    strategy: TruncationStrategy,
) -> None:
    TRUNCATION_REGISTRY[result_type] = strategy


def configure_truncation_limits(
    character_limit: int | None = None,
    list_item_limit: int | None = None,
    list_preview_items: int | None = None,
) -> None:
    for strategy in TRUNCATION_REGISTRY.values():
        if isinstance(strategy, StringFieldTruncation) and character_limit is not None:
            strategy.character_limit = character_limit
        elif isinstance(strategy, ListFieldTruncation):
            if list_item_limit is not None:
                strategy.item_limit = list_item_limit
            if list_preview_items is not None:
                strategy.preview_items = list_preview_items
        elif isinstance(strategy, CompositeTruncation):
            for sub_strategy in strategy.strategies:
                if isinstance(sub_strategy, StringFieldTruncation) and character_limit:
                    sub_strategy.character_limit = character_limit
                elif isinstance(sub_strategy, ListFieldTruncation):
                    if list_item_limit is not None:
                        sub_strategy.item_limit = list_item_limit
                    if list_preview_items is not None:
                        sub_strategy.preview_items = list_preview_items



import os
from asyncio import gather, to_thread
from typing import Final, cast

from anyio import Path as AsyncPath
from python_ripgrep import PySortMode, PySortModeKind, files, search
from rapidfuzz import process

from exponent.core.remote_execution.types import (
    FileAttachment,
    FilePath,
    GetAllTrackedFilesRequest,
    GetAllTrackedFilesResponse,
    GetFileAttachmentRequest,
    GetFileAttachmentResponse,
    GetFileAttachmentsRequest,
    GetFileAttachmentsResponse,
    GetMatchingFilesRequest,
    GetMatchingFilesResponse,
    ListFilesRequest,
    ListFilesResponse,
    RemoteFile,
)
from exponent.core.remote_execution.utils import safe_read_file

MAX_MATCHING_FILES: Final[int] = 10
FILE_NOT_FOUND: Final[str] = "File {} does not exist"
MAX_FILES_TO_WALK: Final[int] = 10_000

GLOB_MAX_COUNT: Final[int] = 1000


class FileCache:
    """A cache of the files in a working directory.

    Args:
        working_directory: The working directory to cache the files from.
    """

    def __init__(self, working_directory: str) -> None:
        self.working_directory = working_directory
        self._cache: list[str] | None = None

    async def get_files(self) -> list[str]:
        """Get the files in the working directory.

        Returns:
            A list of file paths in the working directory.
        """
        if self._cache is None:
            self._cache = await file_walk(self.working_directory)

        return self._cache


async def list_files(list_files_request: ListFilesRequest) -> ListFilesResponse:
    """Get a list of files in the specified directory.

    Args:
        list_files_request: An object containing the directory to list files from.

    Returns:
        A list of RemoteFile objects representing the files in the directory.
    """

    filenames = [
        entry.name async for entry in AsyncPath(list_files_request.directory).iterdir()
    ]

    return ListFilesResponse(
        files=[
            RemoteFile(
                file_path=filename,
                working_directory=list_files_request.directory,
            )
            for filename in filenames
        ],
        correlation_id=list_files_request.correlation_id,
    )


async def get_file_content(
    absolute_path: FilePath, offset: int | None = None, limit: int | None = None
) -> tuple[str, bool]:
    """Get the content of the file at the specified path.

    Args:
        absolute_path: The absolute path to the file.

    Returns:
        A tuple containing the content of the file and a boolean indicating if the file exists.
    """
    file = AsyncPath(absolute_path)
    exists = await file.exists()

    if not exists:
        return FILE_NOT_FOUND.format(absolute_path), False

    if file.is_dir():
        return "File is a directory", True

    content = await safe_read_file(file)

    if offset or limit:
        offset = offset or 0
        limit = limit or -1

        content_lines = content.splitlines()
        content_lines = content_lines[offset:]
        content_lines = content_lines[:limit]

        content = "\n".join(content_lines)

    return content, exists


async def get_file_attachments(
    get_file_attachments_request: GetFileAttachmentsRequest,
    client_working_directory: str,
) -> GetFileAttachmentsResponse:
    """Get the content of the files at the specified paths.

    Args:
        get_file_attachments_request: An object containing the file paths.
        client_working_directory:  The working directory of the client.

    Returns:
        A list of FileAttachment objects containing the content of the files.
    """
    remote_files = get_file_attachments_request.files
    attachments = await gather(
        *[
            get_file_content(
                AsyncPath(client_working_directory) / remote_file.file_path
            )
            for remote_file in remote_files
        ]
    )

    files = [
        FileAttachment(attachment_type="file", file=remote_file, content=content)
        for remote_file, (content, _) in zip(remote_files, attachments)
    ]

    return GetFileAttachmentsResponse(
        correlation_id=get_file_attachments_request.correlation_id,
        file_attachments=files,
    )


async def get_file_attachment(
    get_file_attachment_request: GetFileAttachmentRequest, client_working_directory: str
) -> GetFileAttachmentResponse:
    """Get the content of the file at the specified path.

    Args:
        get_file_attachment_request: An object containing the file path.
        client_working_directory: The working directory of the client.

    Returns:
        A FileAttachment object containing the content of the file.
    """
    file = get_file_attachment_request.file
    absolute_path = await file.resolve(client_working_directory)

    content, exists = await get_file_content(absolute_path)

    return GetFileAttachmentResponse(
        content=content,
        exists=exists,
        file=file,
        correlation_id=get_file_attachment_request.correlation_id,
    )


async def get_matching_files(
    search_term: GetMatchingFilesRequest,
    file_cache: FileCache,
) -> GetMatchingFilesResponse:
    """Get the files that match the search term.

    Args:
        search_term: The search term to match against the files.
        file_cache: A cache of the files in the working directory.

    Returns:
        A list of RemoteFile objects that match the search term.
    """
    # Use rapidfuzz to find the best matching files
    matching_files = await to_thread(
        process.extract,
        search_term.search_term,
        await file_cache.get_files(),
        limit=MAX_MATCHING_FILES,
        score_cutoff=0,
    )

    directory = file_cache.working_directory
    files: list[RemoteFile] = [
        RemoteFile(file_path=file, working_directory=directory)
        for file, _, _ in matching_files
    ]

    return GetMatchingFilesResponse(
        files=files,
        correlation_id=search_term.correlation_id,
    )


async def get_all_tracked_files(
    request: GetAllTrackedFilesRequest,
    working_directory: str,
) -> GetAllTrackedFilesResponse:
    return GetAllTrackedFilesResponse(
        correlation_id=request.correlation_id,
        files=await get_all_non_ignored_files(working_directory),
    )


async def search_files(
    path_str: str,
    file_pattern: str | None,
    regex: str,
    working_directory: str,
) -> list[str]:
    path = AsyncPath(working_directory) / path_str
    path_resolved = await path.resolve()
    globs = [file_pattern] if file_pattern else None

    return await to_thread(
        search,
        patterns=[regex],
        paths=[str(path_resolved)],
        globs=globs,
        after_context=3,
        before_context=5,
        heading=True,
        separator_field_context="|",
        separator_field_match="|",
        separator_context="\n...\n",
    )


async def get_all_file_contents(
    working_directory: str,
) -> list[list[str]]:
    path_resolved = await AsyncPath(working_directory).resolve()

    results = await to_thread(
        search,
        patterns=[".*"],
        paths=[str(path_resolved)],
        globs=["!**/poetry.lock", "!**/pnpm-lock.yaml"],
        heading=True,
        line_number=False,
    )

    result_sizes = [len(result) for result in results]
    total_size = sum(result_sizes)
    batch_size = total_size // 10

    batches = []
    current_batch: list[str] = []
    current_size = 0

    for i, result in enumerate(results):
        if current_size + result_sizes[i] > batch_size:
            batches.append(current_batch)
            current_batch = []
            current_size = 0

        current_batch.append(result)
        current_size += result_sizes[i]

    batches.append(current_batch)

    return batches


async def normalize_files(
    working_directory: str, file_paths: list[FilePath]
) -> list[RemoteFile]:
    """Normalize file paths to be relative to the working directory.

    Args:
        working_directory: The working directory to normalize the file paths against.
        file_paths: A list of file paths to normalize.

    Returns:
        A list of RemoteFile objects with normalized file paths.
    """
    working_path = await AsyncPath(working_directory).resolve()
    normalized_files = []

    for file_path in file_paths:
        path = AsyncPath(file_path)

        if path.is_absolute():
            path = path.relative_to(working_path)

        normalized_files.append(
            RemoteFile(
                file_path=str(path),
                working_directory=working_directory,
            )
        )

    return sorted(normalized_files)


def _format_ignore_globs(ignore_extra: list[str] | None) -> list[str]:
    if ignore_extra is None:
        return []

    return [f"!**/{ignore}" for ignore in ignore_extra]


async def file_walk(
    directory: str,
    ignore_extra: list[str] | None = None,
    max_files: int = MAX_FILES_TO_WALK,
) -> list[str]:
    """
    Walk through a directory and return all file paths, respecting .gitignore and additional ignore patterns.

    Args:
        directory: The directory to walk through
        ignore_extra: Additional directory paths to ignore, follows the gitignore format.
        max_files: The maximal number of files to return

    Returns:
        A list of file paths in the directory.
    """
    working_path = str(await AsyncPath(directory).resolve())

    results: list[str] = await to_thread(
        files,
        patterns=[""],
        paths=[working_path],
        globs=_format_ignore_globs(ignore_extra),
        sort=PySortMode(kind=PySortModeKind.Path),
        max_count=max_files,
    )

    # Create relative paths using os.path functions which handle platform differences
    relative_results = []
    for result in results:
        # Check if the path is inside the working directory
        if os.path.commonpath([working_path, result]) == working_path:
            # Create relative path
            rel_path = os.path.relpath(result, working_path)
            relative_results.append(rel_path)
        else:
            # Fallback to just using the filename
            relative_results.append(os.path.basename(result))

    return relative_results


async def get_all_non_ignored_files(working_directory: str) -> list[RemoteFile]:
    file_paths = await file_walk(working_directory, ignore_extra=DEFAULT_IGNORES)

    return await normalize_files(working_directory, cast(list[FilePath], file_paths))


async def glob(
    path: str,
    glob_pattern: str,
) -> list[str]:
    return await to_thread(
        files,
        patterns=[],
        paths=[path],
        globs=[glob_pattern],
        sort=PySortMode(kind=PySortModeKind.Path),
        max_count=GLOB_MAX_COUNT,
    )


DEFAULT_IGNORES = [
    "**/.git/",
    ".venv/",
    ".mypy_cache",
    ".pytest_cache",
    "node_modules/",
    "venv/",
    ".pyenv",
    "__pycache__",
    ".ipynb_checkpoints",
    ".vercel",
    "__pycache__/",
    "*.py[cod]",
    "*$py.class",
    ".env",
    "*.so",
    ".Python",
    "build/",
    "develop-eggs/",
    "dist/",
    "downloads/",
    "eggs/",
    ".eggs/",
    "lib/",
    "lib64/",
    "parts/",
    "sdist/",
    "var/",
    "wheels/",
    "pip-wheel-metadata/",
    "share/python-wheels/",
    "*.egg-info/",
    ".installed.cfg",
    "*.egg",
    "MANIFEST",
]



import asyncio
import os
import sys
import threading
import time
import webbrowser

import click

from exponent.core.config import Environment, Settings
from exponent.utils.version import get_installed_version


def print_editable_install_forced_prod_warning(settings: Settings) -> None:
    click.secho(
        "Detected local editable install, but this command only works against prod.",
        fg="red",
        bold=True,
    )
    click.secho("Using prod settings:", fg="red", bold=True)
    click.secho("- base_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_url}", fg=(100, 200, 255), bold=False)
    click.secho("- base_api_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.get_base_api_url()}", fg=(100, 200, 255), bold=False)
    click.secho()


def print_editable_install_warning(settings: Settings) -> None:
    click.secho(
        "Detected local editable install, using local URLs", fg="yellow", bold=True
    )
    click.secho("- base_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_url}", fg=(100, 200, 255), bold=False)
    click.secho("- base_api_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.get_base_api_url()}", fg=(100, 200, 255), bold=False)
    click.secho()


def print_exponent_message(base_url: str, chat_uuid: str) -> None:
    version = get_installed_version()
    shell = os.environ.get("SHELL")

    click.echo()
    click.secho(f"△ Indent v{version}", fg=(180, 150, 255), bold=True)
    click.echo()
    click.echo(
        " - Link: " + click.style(f"{base_url}/chats/{chat_uuid}", fg=(100, 200, 255))
    )

    if shell is not None:
        click.echo(f" - Shell: {shell}")


def is_exponent_app_installed() -> bool:
    if sys.platform == "darwin":  # macOS
        return os.path.exists("/Applications/Exponent.app")

    # TODO: Add support for Windows and Linux
    return False


def launch_exponent_browser(
    environment: Environment, base_url: str, chat_uuid: str
) -> None:
    if is_exponent_app_installed() and environment == Environment.production:
        url = f"exponent://chats/{chat_uuid}"
    else:
        url = f"{base_url}/chats/{chat_uuid}"
    webbrowser.open(url)


def start_background_event_loop() -> asyncio.AbstractEventLoop:
    def run_event_loop(loop: asyncio.AbstractEventLoop) -> None:
        asyncio.set_event_loop(loop)
        loop.run_forever()

    loop = asyncio.new_event_loop()
    thread = threading.Thread(target=run_event_loop, args=(loop,), daemon=True)
    thread.start()
    return loop


def read_input(prompt: str) -> str:
    sys.stdout.write(prompt)
    sys.stdout.flush()
    return sys.stdin.readline()


def ask_for_quit_confirmation(program_name: str = "Exponent") -> bool:
    while True:
        try:
            answer = (
                input(f"Do you want to quit {program_name}? [y/\x1b[1mN\x1b[0m]")
                .strip()
                .lower()
            )
            if answer in {"y", "yes"}:
                return True
            elif answer in {"n", "no", ""}:
                return False
        except KeyboardInterrupt:
            print()
            return True
        except EOFError:
            print()
            return True


class Spinner:
    def __init__(self, text: str) -> None:
        self.text = text
        self.task: asyncio.Task[None] | None = None
        self.base_time = time.time()
        self.fg_color: tuple[int, int, int] | None = None
        self.bold = False
        self.animation_chars = "⣷⣯⣟⡿⢿⣻⣽⣾"
        self.animation_speed = 10

    def show(self) -> None:
        if self.task is not None:
            return

        async def spinner(base_time: float) -> None:
            color_start = ""
            if self.fg_color:
                if isinstance(self.fg_color, tuple) and len(self.fg_color) == 3:
                    r, g, b = self.fg_color
                    color_start = f"\x1b[38;2;{r};{g};{b}m"
                elif isinstance(self.fg_color, int):
                    color_start = f"\x1b[{30 + self.fg_color}m"

            bold_start = "\x1b[1m" if self.bold else ""
            style_start = f"{color_start}{bold_start}"
            style_end = "\x1b[0m"

            while True:
                t = time.time() - base_time
                i = round(t * self.animation_speed) % len(self.animation_chars)
                print(
                    f"\r{style_start}{self.animation_chars[i]} {self.text}{style_end}",
                    end="",
                )
                await asyncio.sleep(0.1)

        self.task = asyncio.get_event_loop().create_task(spinner(self.base_time))

    def hide(self) -> None:
        if self.task is None:
            return

        self.task.cancel()
        self.task = None
        print("\r\x1b[0m\x1b[2K", end="")
        sys.stdout.flush()


class ThinkingSpinner(Spinner):
    def __init__(
        self,
        fg_color: tuple[int, int, int] | None = None,
        text: str = "Exponent is thinking",
    ) -> None:
        super().__init__(text)
        self.fg_color = fg_color
        self.bold = True
        self.animation_chars = "⣾⣽⣻⢿⡿⣟⣯⣷"  # More subtle animation
        self.animation_speed = 8  # Medium speed animation
        self.start_time = time.time()  # Track when thinking started

    def show(self) -> None:
        if self.task is not None:
            return

        async def spinner(base_time: float) -> None:
            color_start = ""
            if self.fg_color:
                if isinstance(self.fg_color, tuple) and len(self.fg_color) == 3:
                    r, g, b = self.fg_color
                    color_start = f"\x1b[38;2;{r};{g};{b}m"
                elif isinstance(self.fg_color, int):
                    color_start = f"\x1b[{30 + self.fg_color}m"

            bold_start = "\x1b[1m" if self.bold else ""
            style_start = f"{color_start}{bold_start}"
            style_end = "\x1b[0m"

            while True:
                t = time.time() - base_time
                i = round(t * self.animation_speed) % len(self.animation_chars)

                # Calculate elapsed seconds
                elapsed = time.time() - self.start_time
                if elapsed < 60:
                    timer = f"{int(elapsed)}s"
                else:
                    mins = int(elapsed // 60)
                    secs = int(elapsed % 60)
                    timer = f"{mins}m {secs}s"

                print(
                    f"\r{style_start}{self.animation_chars[i]} {self.text} ({timer}){style_end}",
                    end="",
                )
                await asyncio.sleep(0.1)

        self.task = asyncio.get_event_loop().create_task(spinner(self.base_time))


class ConnectionTracker:
    def __init__(self) -> None:
        self.connected = True
        self.queue: asyncio.Queue[bool] = asyncio.Queue()

    def is_connected(self) -> bool:
        while True:
            try:
                self.connected = self.queue.get_nowait()
            except asyncio.QueueEmpty:
                break

        return self.connected

    async def wait_for_reconnection(self) -> None:
        if not self.is_connected():
            assert await self.queue.get()
            self.connected = True

    async def set_connected(self, connected: bool) -> None:
        await self.queue.put(connected)

    async def next_change(self) -> bool:
        return await self.queue.get()


def get_short_git_commit_hash(commit_hash: str) -> str:
    return commit_hash[:8]



import uuid
from collections.abc import Callable
from pathlib import Path
from time import time

from anyio import Path as AsyncPath

from exponent.core.remote_execution import files
from exponent.core.remote_execution.cli_rpc_types import (
    BashToolInput,
    BashToolResult,
    ErrorToolResult,
    GlobToolInput,
    GlobToolResult,
    GrepToolInput,
    GrepToolResult,
    ListToolInput,
    ListToolResult,
    ReadToolInput,
    ReadToolResult,
    ToolInputType,
    ToolResultType,
    WriteToolInput,
    WriteToolResult,
)
from exponent.core.remote_execution.code_execution import execute_code
from exponent.core.remote_execution.file_write import execute_full_file_rewrite
from exponent.core.remote_execution.truncation import truncate_tool_result
from exponent.core.remote_execution.types import CodeExecutionRequest
from exponent.core.remote_execution.utils import (
    assert_unreachable,
    safe_read_file,
)

GREP_MAX_RESULTS = 100


async def execute_tool(
    tool_input: ToolInputType, working_directory: str
) -> ToolResultType:
    if isinstance(tool_input, ReadToolInput):
        return await execute_read_file(tool_input, working_directory)
    elif isinstance(tool_input, WriteToolInput):
        return await execute_write_file(tool_input, working_directory)
    elif isinstance(tool_input, ListToolInput):
        return await execute_list_files(tool_input, working_directory)
    elif isinstance(tool_input, GlobToolInput):
        return await execute_glob_files(tool_input, working_directory)
    elif isinstance(tool_input, GrepToolInput):
        return await execute_grep_files(tool_input, working_directory)
    elif isinstance(tool_input, BashToolInput):
        raise ValueError("Bash tool input should be handled by execute_bash_tool")
    else:
        assert_unreachable(tool_input)


def truncate_result[T: ToolResultType](tool_result: T) -> T:
    return truncate_tool_result(tool_result)


async def execute_read_file(  # noqa: PLR0911
    tool_input: ReadToolInput, working_directory: str
) -> ReadToolResult | ErrorToolResult:
    # Validate absolute path requirement
    if not tool_input.file_path.startswith("/"):
        return ErrorToolResult(
            error_message=f"File path must be absolute, got relative path: {tool_input.file_path}"
        )

    # Validate offset and limit
    offset = tool_input.offset if tool_input.offset is not None else 0
    limit = tool_input.limit if tool_input.limit is not None else 2000

    if offset < 0:
        return ErrorToolResult(
            error_message=f"Offset must be non-negative, got: {offset}"
        )

    if limit <= 0:
        return ErrorToolResult(error_message=f"Limit must be positive, got: {limit}")

    file = AsyncPath(working_directory, tool_input.file_path)

    try:
        exists = await file.exists()
    except (OSError, PermissionError) as e:
        return ErrorToolResult(error_message=f"Cannot access file: {e!s}")

    if not exists:
        return ErrorToolResult(
            error_message="File not found",
        )

    try:
        if await file.is_dir():
            return ErrorToolResult(
                error_message=f"{await file.absolute()} is a directory",
            )
    except (OSError, PermissionError) as e:
        return ErrorToolResult(error_message=f"Cannot check file type: {e!s}")

    try:
        content = await safe_read_file(file)
    except PermissionError:
        return ErrorToolResult(
            error_message=f"Permission denied: cannot read {tool_input.file_path}"
        )
    except UnicodeDecodeError:
        return ErrorToolResult(
            error_message="File appears to be binary or has invalid text encoding"
        )
    except Exception as e:  # noqa: BLE001
        return ErrorToolResult(error_message=f"Error reading file: {e!s}")

    # Handle empty files
    if not content:
        return ReadToolResult(
            content="",
            num_lines=0,
            start_line=0,
            total_lines=0,
        )

    content_lines = content.splitlines(keepends=True)
    total_lines = len(content_lines)

    # Handle offset beyond file length
    if offset >= total_lines:
        return ReadToolResult(
            content="",
            num_lines=0,
            start_line=offset,
            total_lines=total_lines,
        )

    # Apply offset and limit
    content_lines = content_lines[offset : offset + limit]

    # Apply character-level truncation at line boundaries to ensure consistency
    # This ensures the content field and num_lines field remain in sync
    CHARACTER_LIMIT = 90_000  # Match the limit in truncation.py

    # Join lines and check total size
    final_content = "".join(content_lines)

    if len(final_content) > CHARACTER_LIMIT:
        # Truncate at line boundaries to stay under the limit
        truncated_lines = []
        current_size = 0
        truncation_message = "\n[Content truncated due to size limit]"
        truncation_size = len(truncation_message)
        lines_included = 0

        for line in content_lines:
            # Check if adding this line would exceed the limit (accounting for truncation message)
            if current_size + len(line) + truncation_size > CHARACTER_LIMIT:
                final_content = "".join(truncated_lines) + truncation_message
                break
            truncated_lines.append(line)
            current_size += len(line)
            lines_included += 1
        else:
            # All lines fit (shouldn't happen if we got here, but be safe)
            final_content = "".join(truncated_lines)
            lines_included = len(content_lines)

        num_lines = lines_included
    else:
        num_lines = len(content_lines)

    return ReadToolResult(
        content=final_content,
        num_lines=num_lines,
        start_line=offset,
        total_lines=total_lines,
    )


async def execute_write_file(
    tool_input: WriteToolInput, working_directory: str
) -> WriteToolResult:
    file_path = tool_input.file_path
    path = Path(working_directory, file_path)
    result = await execute_full_file_rewrite(
        path, tool_input.content, working_directory
    )
    return WriteToolResult(message=result)


async def execute_list_files(
    tool_input: ListToolInput, working_directory: str
) -> ListToolResult | ErrorToolResult:
    path = AsyncPath(tool_input.path)

    try:
        exists = await path.exists()
    except (OSError, PermissionError) as e:
        return ErrorToolResult(error_message=f"Cannot access path: {e!s}")

    if not exists:
        return ErrorToolResult(error_message=f"Directory not found: {tool_input.path}")

    try:
        is_dir = await path.is_dir()
    except (OSError, PermissionError) as e:
        return ErrorToolResult(
            error_message=f"Cannot check if path is directory: {e!s}"
        )

    if not is_dir:
        return ErrorToolResult(
            error_message=f"Path is not a directory: {tool_input.path}"
        )

    try:
        filenames = [entry.name async for entry in path.iterdir()]
    except (OSError, PermissionError) as e:
        return ErrorToolResult(error_message=f"Cannot list directory contents: {e!s}")

    return ListToolResult(
        files=[filename for filename in filenames],
    )


async def execute_glob_files(
    tool_input: GlobToolInput, working_directory: str
) -> GlobToolResult:
    # async timer
    start_time = time()
    results = await files.glob(
        path=working_directory if tool_input.path is None else tool_input.path,
        glob_pattern=tool_input.pattern,
    )
    duration_ms = int((time() - start_time) * 1000)
    return GlobToolResult(
        filenames=results,
        duration_ms=duration_ms,
        num_files=len(results),
        truncated=len(results) >= files.GLOB_MAX_COUNT,
    )


async def execute_grep_files(
    tool_input: GrepToolInput, working_directory: str
) -> GrepToolResult:
    results = await files.search_files(
        path_str=working_directory if tool_input.path is None else tool_input.path,
        file_pattern=tool_input.include,
        regex=tool_input.pattern,
        working_directory=working_directory,
    )
    return GrepToolResult(
        matches=results[:GREP_MAX_RESULTS],
        truncated=bool(len(results) > GREP_MAX_RESULTS),
    )


async def execute_bash_tool(
    tool_input: BashToolInput, working_directory: str, should_halt: Callable[[], bool]
) -> BashToolResult:
    start_time = time()
    result = await execute_code(
        CodeExecutionRequest(
            language="shell",
            content=tool_input.command,
            timeout=120 if tool_input.timeout is None else tool_input.timeout,
            correlation_id=str(uuid.uuid4()),
        ),
        working_directory=working_directory,
        session=None,  # type: ignore
        should_halt=should_halt,
    )
    return BashToolResult(
        shell_output=result.content,
        exit_code=result.exit_code,
        duration_ms=int((time() - start_time) * 1000),
        timed_out=result.cancelled_for_timeout,
        stopped_by_user=result.halted,
    )



import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Optional

import httpx
from httpx import AsyncClient, Request, Response

from exponent.core.remote_execution.exceptions import ExponentError
from exponent.core.remote_execution.languages.python_execution import Kernel
from exponent.core.remote_execution.utils import format_error_log

if TYPE_CHECKING:
    from exponent.core.config import Settings

logger = logging.getLogger(__name__)


class SessionLog:
    def __init__(self) -> None:
        self.log_buffer: list[str] = []
        self.max_size = 5

    def append_log(self, log: str) -> None:
        self.log_buffer.append(log)
        self.log_buffer = self.log_buffer[-self.max_size :]

    def get_logs(self) -> list[str]:
        return self.log_buffer

    async def log_request(self, request: Request) -> None:
        self.append_log(f"Request: {request.method} {request.url}")

    async def log_response(self, response: Response) -> None:
        request = response.request
        await response.aread()
        self.append_log(
            f"Response for request: {request.method} {request.url}\n"
            f"Response: {response.status_code}, {response.text}"
        )


class RemoteExecutionClientSession:
    def __init__(
        self, working_directory: str, base_url: str, base_ws_url: str, api_key: str
    ):
        self.chat_uuid: str | None = None

        self.working_directory = working_directory
        self.kernel = Kernel(working_directory=working_directory)
        self.api_log = SessionLog()

        self.api_client = AsyncClient(
            base_url=base_url,
            headers={"API-KEY": api_key},
            event_hooks={
                "request": [self.api_log.log_request],
                "response": [self.api_log.log_response],
            },
        )

        self.ws_client = AsyncClient(
            base_url=base_ws_url,
            headers={"API-KEY": api_key},
            event_hooks={
                "request": [self.api_log.log_request],
                "response": [self.api_log.log_response],
            },
        )

    def set_chat_uuid(self, chat_uuid: str) -> None:
        self.chat_uuid = chat_uuid


async def send_exception_log(
    exc: Exception,
    session: RemoteExecutionClientSession | None = None,
    settings: Optional["Settings"] = None,
) -> None:
    error_log = format_error_log(
        exc=exc,
        chat_uuid=session.chat_uuid if session else None,
        attachment_lines=session.api_log.get_logs() if session else None,
    )

    if session:
        api_client = session.api_client
    elif settings:
        api_key = settings.api_key
        if not api_key:
            raise ValueError("No API key provided")
        api_client = AsyncClient(
            base_url=settings.get_base_api_url(),
            headers={"API-KEY": api_key},
        )
    else:
        raise ValueError("No session or settings provided")

    if not error_log:
        return
    try:
        await api_client.post(
            "/api/remote_execution/log_error",
            content=error_log.model_dump_json(),
            timeout=60,
        )
    except httpx.ConnectError:
        logger.info("Failed to send error log")


@asynccontextmanager
async def get_session(
    working_directory: str,
    base_url: str,
    base_ws_url: str,
    api_key: str,
) -> AsyncGenerator[RemoteExecutionClientSession, None]:
    session = RemoteExecutionClientSession(
        working_directory, base_url, base_ws_url, api_key
    )
    try:
        yield session
    except Exception as exc:  # noqa: BLE001
        await send_exception_log(exc, session=session, settings=None)
        raise ExponentError(str(exc))
    finally:
        session.kernel.close()
        await session.api_client.aclose()



import asyncio
import os
import platform
import shutil
from collections.abc import Callable
from dataclasses import dataclass

STDOUT_FD = 1
STDERR_FD = 2
MAX_TIMEOUT = 300


"""
Deprecated, use shell_streaming.py
"""


@dataclass
class ShellExecutionResult:
    output: str
    cancelled_for_timeout: bool
    exit_code: int | None
    halted: bool = False


def get_rc_file_source_command(shell_path: str) -> str:
    """
    Returns a command to source the user's shell rc file
    Login profiles are already sourced via the -l flag
    """
    # On Windows, shell behavior is different
    if platform.system() == "Windows":
        return ""  # Windows shells don't typically use rc files in the same way

    shell_name = os.path.basename(shell_path)
    home_dir = os.path.expanduser("~")

    if shell_name == "zsh":
        zshrc = os.path.join(home_dir, ".zshrc")
        if os.path.exists(zshrc):
            return f"source {zshrc} 2>/dev/null || true; "
    elif shell_name == "bash":
        bashrc = os.path.join(home_dir, ".bashrc")
        if os.path.exists(bashrc):
            return f"source {bashrc} 2>/dev/null || true; "

    return ""  # No rc file found or unsupported shell


def is_native_windows() -> bool:
    return os.name == "nt" and "WSL_DISTRO_NAME" not in os.environ


async def execute_shell(
    code: str,
    working_directory: str,
    timeout: int,
    should_halt: Callable[[], bool] | None = None,
) -> ShellExecutionResult:
    timeout = min(timeout, MAX_TIMEOUT)

    shell_path = os.environ.get("SHELL") or shutil.which("bash") or shutil.which("sh")

    if not shell_path:
        process = await asyncio.create_subprocess_shell(
            code,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=working_directory,
        )
    else:
        # Add rc file sourcing to the command
        rc_source_cmd = get_rc_file_source_command(shell_path)
        full_command = f"{rc_source_cmd}{code}"

        process = await asyncio.create_subprocess_exec(
            shell_path,
            "-l",
            "-c",
            full_command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=working_directory,
        )

    exit_code = None
    output: list[tuple[int, str]] = []
    halted = False
    assert process.stdout
    assert process.stderr

    stdout_capture_task = asyncio.create_task(
        capture(process.stdout, STDOUT_FD, output)
    )
    stderr_capture_task = asyncio.create_task(
        capture(process.stderr, STDERR_FD, output)
    )

    async def capture_until_exit() -> int:
        nonlocal halted
        while True:
            if should_halt and should_halt():
                process.kill()
                halted = True
                break

            if process.returncode is not None:
                break

            await asyncio.sleep(0.1)

        return await process.wait()

    try:
        exit_code = await asyncio.wait_for(capture_until_exit(), timeout)
    except TimeoutError:
        process.kill()
    except Exception:
        raise
    finally:
        # Wait for capture tasks to complete naturally after streams are closed
        await asyncio.gather(
            stdout_capture_task, stderr_capture_task, return_exceptions=True
        )

    formatted_output = "".join([chunk for (_, chunk) in output]).strip() + "\n\n"

    return ShellExecutionResult(
        output=formatted_output,
        cancelled_for_timeout=exit_code is None and not halted,
        exit_code=exit_code,
        halted=halted,
    )


async def capture(
    stream: asyncio.StreamReader, fd: int, output: list[tuple[int, str]]
) -> None:
    while True:
        data = await stream.read(4096)
        if not data:
            break

        chunk = data.decode(errors="replace")
        output.append((fd, chunk))



CREATE_GITHUB_CONFIG_MUTATION: str = """
    mutation CreateGithubConfig(
        $githubPat: String!,
    ) {
        createGithubConfig(
            githubPat: $githubPat
        ) {
            __typename
            ... on GithubConfig {
                githubConfigUuid
                githubPat
            }
        }
    }
"""

CHECK_GITHUB_CONFIG_VALIDITY_QUERY: str = """
    query CheckGithubConfigValidity {
        checkGithubConfigValidity {
            __typename
            ... on GithubConfigValidityResult {
                isValid
                message
            }
            ... on Error {
                message
            }
        }
    }
"""

REPOS_FOR_GITHUB_CONFIG_QUERY: str = """
    query ReposForGithubConfig {
        reposForGithubConfig {
            __typename
            ... on GithubConfigRepos {
                repos {
                    id
                    name
                    fullName
                    private
                    owner
                    description
                }
                orgs {
                    login
                    id
                    url
                }
            }
            ... on Error {
                message
            }
        }
    }
"""



AUTHENTICATED_USER_SUBSCRIPTION = """
    subscription {
            testAuthenticatedUser {
                __typename
                ... on UnauthenticatedError {
                    message
                }
                ...on Error {
                    message
                }
                ... on User {
                    userUuid
                }
            }
        }
"""

INDENT_EVENTS_SUBSCRIPTION = """
  subscription ChatEvents(
    $prompt: Prompt
    $chatUuid: String!
    $parentUuid: String
    $model: ModelName!
    $strategyNameOverride: StrategyName
    $depthLimit: Int!
    $requireConfirmation: Boolean
    $readOnly: Boolean
    $enableThinking: Boolean
  ) {
    indentChat(
      chatInput: {
        prompt: $prompt
      }
      parentUuid: $parentUuid
      chatConfig: {
        chatUuid: $chatUuid
        model: $model
        requireConfirmation: $requireConfirmation
        readOnly: $readOnly
        strategyNameOverride: $strategyNameOverride
        depthLimit: $depthLimit
        enableThinking: $enableThinking
      }
    ) {
      __typename
      ...on Error {
        message
      }
      ...on UnauthenticatedError {
        message
      }
      ...on UserEvent {
        uuid
        parentUuid
        chatId
        isSidechain
        version
        createdAt
        sidechainRootUuid
        synthetic
        messageData: message {
          __typename
          ... on TextMessage {
            text
          }
          ... on ToolResultMessage {
            messageId
            toolUseId
            text
            resultData {
              ... on BashToolResult {
                shellOutput
                exitCode
              }
              ... on ReadToolResult {
                content
              }
            }
          }
          ... on PartialToolResultMessage {
            messageId
            toolUseId
            text
            resultData {
              ... on PartialToolResult {
                __typename
              }
            }
          }
        }
      }
      ...on AssistantEvent {
        uuid
        parentUuid
        chatId
        isSidechain
        version
        createdAt
        sidechainRootUuid
        synthetic
        messageData: message {
          __typename
          ... on TextMessage {
            text
          }
          ... on ToolCallMessage {
            messageId
            toolUseId
            toolName
            toolInput {
              ... on BashToolInput {
                command
              }
              ... on ReadToolInput {
                filePath
              }
            }
          }
        }
      }
      ...on SystemEvent {
        uuid
        parentUuid
        chatId
        isSidechain
        version
        createdAt
        sidechainRootUuid
        messageData: message {
          __typename
          ... on ToolCallMessage {
            messageId
            toolUseId
            toolName
            toolInput {
              ... on BashToolInput {
                command
              }
              ... on ReadToolInput {
                filePath
              }
            }
          }
          ... on ToolResultMessage {
            messageId
            toolUseId
            text
            resultData {
              ... on BashToolResult {
                shellOutput
                exitCode
              }
              ... on ReadToolResult {
                content
              }
            }
          }
          ... on ToolExecutionStatusMessage {
            executionStatus: status
          }
          ... on ToolPermissionStatusMessage {
            permissionStatus: status
          }
        }
      }
    }
  }
"""


CONFIRM_AND_CONTINUE_SUBSCRIPTION = """
  subscription ChatEvents(
    $requestUuid: String!
    $chatUuid: String!
    $model: ModelName!
    $strategyNameOverride: StrategyName
    $depthLimit: Int!
    $requireConfirmation: Boolean
    $readOnly: Boolean
    $enableThinking: Boolean
  ) {
    confirmAndContinue(
      requestEventUuid: $requestUuid
      chatConfig: {
        chatUuid: $chatUuid
        model: $model
        requireConfirmation: $requireConfirmation
        readOnly: $readOnly
        strategyNameOverride: $strategyNameOverride
        depthLimit: $depthLimit
        enableThinking: $enableThinking
      }
    ) {
      __typename
      ...on Error {
        message
      }
      ...on UnauthenticatedError {
        message
      }
      ...on UserEvent {
        uuid
        parentUuid
        chatId
        isSidechain
        version
        createdAt
        sidechainRootUuid
        synthetic
        messageData: message {
          __typename
          ... on TextMessage {
            text
          }
          ... on ToolResultMessage {
            messageId
            toolUseId
            text
            resultData {
              ... on BashToolResult {
                shellOutput
                exitCode
              }
              ... on ReadToolResult {
                content
              }
            }
          }
          ... on PartialToolResultMessage {
            messageId
            toolUseId
            text
            resultData {
              ... on PartialToolResult {
                __typename
              }
            }
          }
        }
      }
      ...on AssistantEvent {
        uuid
        parentUuid
        chatId
        isSidechain
        version
        createdAt
        sidechainRootUuid
        synthetic
        messageData: message {
          __typename
          ... on TextMessage {
            text
          }
          ... on ToolCallMessage {
            messageId
            toolUseId
            toolName
            toolInput {
              ... on BashToolInput {
                command
              }
              ... on ReadToolInput {
                filePath
              }
            }
          }
        }
      }
      ...on SystemEvent {
        uuid
        parentUuid
        chatId
        isSidechain
        version
        createdAt
        sidechainRootUuid
        messageData: message {
          __typename
          ... on ToolCallMessage {
            messageId
            toolUseId
            toolName
            toolInput {
              ... on BashToolInput {
                command
              }
              ... on ReadToolInput {
                filePath
              }
            }
          }
          ... on ToolResultMessage {
            messageId
            toolUseId
            text
            resultData {
              ... on BashToolResult {
                shellOutput
                exitCode
              }
              ... on ReadToolResult {
                content
              }
            }
          }
          ... on ToolExecutionStatusMessage {
            executionStatus: status
          }
          ... on ToolPermissionStatusMessage {
            permissionStatus: status
          }
        }
      }
    }
  }
"""


INDENT_CHAT_EVENT_STREAM_SUBSCRIPTION = """
  subscription IndentChatEventStream(
    $chatUuid: String!
    $knownFullEventUuids: [String!]
  ) {
    indentChatEventStream(
      chatUuid: $chatUuid
      knownFullEventUuids: $knownFullEventUuids
    ) {
      __typename
      ...on Error {
        message
      }
      ...on UnauthenticatedError {
        message
      }
      ...on UserEvent {
        uuid
        parentUuid
        chatId
        isSidechain
        version
        createdAt
        sidechainRootUuid
        synthetic
        messageData: message {
          __typename
          ... on TextMessage {
            text
          }
          ... on ToolResultMessage {
            messageId
            toolUseId
            text
            resultData {
              ... on BashToolResult {
                shellOutput
                exitCode
              }
              ... on ReadToolResult {
                content
              }
            }
          }
          ... on PartialToolResultMessage {
            messageId
            toolUseId
            text
            resultData {
              ... on PartialToolResult {
                __typename
              }
            }
          }
        }
      }
      ...on AssistantEvent {
        uuid
        parentUuid
        chatId
        isSidechain
        version
        createdAt
        sidechainRootUuid
        synthetic
        messageData: message {
          __typename
          ... on TextMessage {
            text
          }
          ... on ToolCallMessage {
            messageId
            toolUseId
            toolName
            toolInput {
              ... on BashToolInput {
                command
              }
              ... on ReadToolInput {
                filePath
              }
            }
          }
        }
      }
      ...on SystemEvent {
        uuid
        parentUuid
        chatId
        isSidechain
        version
        createdAt
        sidechainRootUuid
        messageData: message {
          __typename
          ... on ToolCallMessage {
            messageId
            toolUseId
            toolName
            toolInput {
              ... on BashToolInput {
                command
              }
              ... on ReadToolInput {
                filePath
              }
            }
          }
          ... on ToolResultMessage {
            messageId
            toolUseId
            text
            resultData {
              ... on BashToolResult {
                shellOutput
                exitCode
              }
              ... on ReadToolResult {
                content
              }
            }
          }
          ... on ToolExecutionStatusMessage {
            executionStatus: status
          }
          ... on ToolPermissionStatusMessage {
            permissionStatus: status
          }
        }
      }
    }
  }
"""



import asyncio
import logging
import queue
import re
import sys
import threading
import time
from collections.abc import AsyncGenerator, Callable
from typing import Any

from jupyter_client.client import KernelClient
from jupyter_client.manager import KernelManager

from exponent.core.remote_execution.languages.types import (
    PythonExecutionResult,
    StreamedOutputPiece,
)
from exponent.core.remote_execution.types import PythonEnvInfo

logger = logging.getLogger(__name__)


class IOChannelHandler:
    ESCAPE_SEQUENCE = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")

    def __init__(self, user_interrupted: Callable[[], bool] | None = None) -> None:
        self.output_buffer: queue.Queue[str] = queue.Queue()
        self.user_interrupted = user_interrupted

    def add_message(self, message: dict[str, Any]) -> None:
        logger.debug(f"Jupyter kernel message received: {message}")
        output = None
        if message["msg_type"] == "stream":
            output = message["content"]["text"]
        elif message["msg_type"] == "error":
            raw_content = "\n".join(message["content"]["traceback"])
            content = self.ESCAPE_SEQUENCE.sub("", raw_content)
            output = content
        if output:
            self.output_buffer.put(output)

    @staticmethod
    def is_idle(message: dict[str, Any]) -> bool:
        return bool(
            message["header"]["msg_type"] == "status"
            and message["content"]["execution_state"] == "idle"
        )


class Kernel:
    def __init__(self, working_directory: str) -> None:
        self._manager: KernelManager | None = None
        self._client: KernelClient | None = None
        self.io_handler: IOChannelHandler = IOChannelHandler()
        self.working_directory = working_directory
        self.interrupted_by_user: bool = False

    @property
    def manager(self) -> KernelManager:
        if not self._manager:
            self._manager = KernelManager(kernel_name="python3")
            self._manager.start_kernel(cwd=self.working_directory)
        return self._manager

    @property
    def client(self) -> KernelClient:
        if not self._client:
            self._client = self.manager.client()

            while not self._client.is_alive():
                time.sleep(0.1)

            self._client.start_channels()
        return self._client

    async def wait_for_ready(self, timeout: int = 5) -> None:
        manager = self.manager
        start_time = time.time()
        while not manager.is_alive():
            if time.time() - start_time > timeout:
                raise Exception("Kernel took too long to start")
            await asyncio.sleep(0.05)
        await asyncio.sleep(0.5)

    def _clear_channels(self) -> None:
        """Clear all pending messages from kernel channels."""
        # First clear shell and control channels
        channels = [
            self.client.shell_channel,
            self.client.control_channel,
        ]
        for channel in channels:
            try:
                while True:
                    channel.get_msg(timeout=0.1)
            except queue.Empty:
                continue

        # Then process iopub channel until we get an idle state
        iterations = 0
        while True:
            try:
                msg = self.client.iopub_channel.get_msg(timeout=0.1)
                self.io_handler.add_message(msg)
                if self.io_handler.is_idle(msg):
                    break
            except queue.Empty:
                iterations += 1
                if iterations > 10:
                    logger.info("Kernel took too long to become idle")
                    break

    def iopub_listener(self, client: KernelClient) -> None:
        while True:
            try:
                if (
                    self.io_handler.user_interrupted
                    and self.io_handler.user_interrupted()
                ):
                    logger.info("External halt signal received")
                    self.manager.interrupt_kernel()
                    self.interrupted_by_user = True

                    # Wait for kernel to push any final output
                    time.sleep(0.5)

                    # Clear all channels to reset kernel state
                    self._clear_channels()

                    break

                try:
                    msg = client.iopub_channel.get_msg(timeout=1)
                    logger.debug(f"Received message from kernel: {msg}")
                    self.io_handler.add_message(msg)

                    if self.io_handler.is_idle(msg):
                        logger.debug("Kernel is idle.")
                        break
                except queue.Empty:
                    continue

            except Exception as e:  # noqa: BLE001 - TODO: Deep audit potential exceptions
                logger.info(f"Error getting message from kernel: {e}")
                break

    # Deprecated, use execute_code_streaming
    async def execute_code(self, code: str) -> str:
        async for result in self.execute_code_streaming(code):
            if isinstance(result, PythonExecutionResult):
                return result.output
        # should be unreachable
        raise Exception("No result from kernel")

    async def execute_code_streaming(
        self, code: str, user_interrupted: Callable[[], bool] | None = None
    ) -> AsyncGenerator[StreamedOutputPiece | PythonExecutionResult, None]:
        await self.wait_for_ready()
        self.interrupted_by_user = False

        self.io_handler = IOChannelHandler(user_interrupted=user_interrupted)

        client = self.client
        client.connect_iopub()
        iopub_thread = threading.Thread(target=self.iopub_listener, args=(client,))
        logger.info("Starting IO listener thread.")
        iopub_thread.start()

        logger.info("Executing code in kernel.")
        client.execute(code)

        stopping = False

        results = []
        while True:
            stopping = not iopub_thread.is_alive()

            while not self.io_handler.output_buffer.empty():
                output = self.io_handler.output_buffer.get()
                logger.info("Execution output: %s", output)
                yield StreamedOutputPiece(content=output)
                results.append(output)

            if stopping:
                break

            await asyncio.sleep(0.05)

        # Wait for thread to fully exit
        iopub_thread.join(timeout=1.0)
        yield PythonExecutionResult(
            output="".join(results), halted=self.interrupted_by_user
        )

    def close(self) -> None:
        if self._client:
            self._client.stop_channels()
            self._client = None
        if self._manager:
            self._manager.shutdown_kernel()
            self._manager = None


async def execute_python(code: str, kernel: Kernel) -> str:
    return await kernel.execute_code(code)


async def execute_python_streaming(
    code: str, kernel: Kernel, user_interrupted: Callable[[], bool] | None = None
) -> AsyncGenerator[StreamedOutputPiece | PythonExecutionResult, None]:
    async for result in kernel.execute_code_streaming(
        code, user_interrupted=user_interrupted
    ):
        yield result


### Environment Helpers


def get_python_env_info() -> PythonEnvInfo:
    return PythonEnvInfo(
        interpreter_path=get_active_python_interpreter_path(),
        interpreter_version=get_active_python_interpreter_version(),
    )


def get_active_python_interpreter_path() -> str | None:
    return sys.executable


def get_active_python_interpreter_version() -> str | None:
    version = sys.version

    match = re.search(r"(\d+\.\d+\.\d+).*", version)

    if match:
        return match.group(1)

    return None



import getpass
import os
import platform

from anyio import Path as AsyncPath

from exponent.core.remote_execution.git import get_git_info
from exponent.core.remote_execution.languages import python_execution
from exponent.core.remote_execution.types import (
    SystemContextRequest,
    SystemContextResponse,
    SystemInfo,
)
from exponent.core.remote_execution.utils import safe_read_file

EXPONENT_TXT_FILENAMES = [
    "exponent.txt",
]


async def get_system_context(
    request: SystemContextRequest, working_directory: str
) -> SystemContextResponse:
    return SystemContextResponse(
        correlation_id=request.correlation_id,
        exponent_txt=await _read_exponent_txt(working_directory),
        system_info=await get_system_info(working_directory),
    )


async def get_system_info(working_directory: str) -> SystemInfo:
    return SystemInfo(
        name=getpass.getuser(),
        cwd=working_directory,
        os=platform.system(),
        shell=_get_user_shell(),
        git=await get_git_info(working_directory),
        python_env=python_execution.get_python_env_info(),
    )


async def _read_exponent_txt(working_directory: str) -> str | None:
    for filename in EXPONENT_TXT_FILENAMES:
        file_path = AsyncPath(os.path.join(working_directory, filename.lower()))
        exists = await file_path.exists()

        if exists:
            return await safe_read_file(file_path)

    return None


def _get_user_shell() -> str:
    return os.environ.get("SHELL", "bash")



import asyncio
import sys

import click

from exponent.commands.common import (
    check_inside_git_repo,
    check_running_from_home_directory,
    check_ssl,
    create_cloud_chat,
    redirect_to_login,
    start_chat_turn,
)
from exponent.commands.settings import use_settings
from exponent.commands.types import exponent_cli_group
from exponent.commands.utils import (
    launch_exponent_browser,
    print_exponent_message,
)
from exponent.core.config import Settings
from exponent.utils.version import check_exponent_version_and_upgrade


@exponent_cli_group(hidden=True)
def cloud_cli() -> None:
    pass


@cloud_cli.command(hidden=True)
@click.option(
    "--cloud-config-id",
    help="ID of an existing cloud config to reconnect",
    required=True,
)
@click.option(
    "--prompt",
    help="Prompt to kick off the cloud session.",
    required=True,
)
@click.option(
    "--background",
    "-b",
    help="Start the cloud session without launching the Exponent UI",
    is_flag=True,
    default=False,
)
@use_settings
def cloud(
    settings: Settings,
    cloud_config_id: str,
    prompt: str,
    background: bool,
) -> None:
    check_exponent_version_and_upgrade(settings)

    if not settings.api_key:
        redirect_to_login(settings)
        return

    loop = asyncio.get_event_loop()

    check_running_from_home_directory()
    loop.run_until_complete(check_inside_git_repo(settings))
    check_ssl()

    api_key = settings.api_key
    base_url = settings.base_url
    base_api_url = settings.get_base_api_url()
    base_ws_url = settings.get_base_ws_url()

    chat_uuid = loop.run_until_complete(
        create_cloud_chat(api_key, base_api_url, base_ws_url, cloud_config_id)
    )

    if chat_uuid is None:
        sys.exit(1)

    loop.run_until_complete(
        start_chat_turn(api_key, base_api_url, base_ws_url, chat_uuid, prompt)
    )

    print_exponent_message(base_url, chat_uuid)

    if not background:
        launch_exponent_browser(settings.environment, base_url, chat_uuid)



import asyncio
import json
from pathlib import Path
import subprocess
import sys
from uuid import uuid4
import webbrowser

import click
from git import GitCommandError, Repo

from exponent.commands.common import (
    redirect_to_login,
    verify_gh_app_installation,
)
from exponent.commands.settings import use_settings
from exponent.commands.types import exponent_cli_group
from exponent.core.config import Settings
from exponent.core.remote_execution.git import get_git_info


@exponent_cli_group()
def github_app_cli() -> None:
    """Run AI-powered chat sessions."""
    pass


@github_app_cli.command()
@use_settings
def install_github_app(
    settings: Settings,
) -> None:
    """Start or reconnect to an Exponent session."""
    if not settings.api_key:
        redirect_to_login(settings)
        return

    loop = asyncio.get_event_loop()

    api_key = settings.api_key
    base_api_url = settings.get_base_api_url()
    base_ws_url = settings.get_base_ws_url()

    git_info = asyncio.run(get_git_info("."))
    if not git_info:
        raise RuntimeError("Not running inside of valid git repository")

    install_url = "https://github.com/apps/indent-com/installations/new"
    webbrowser.open(install_url)

    click.confirm(
        "Press enter once you've installed the github app.",
        default=True,
        abort=True,
        prompt_suffix="",
    )

    click.secho("Verifying installation...", fg="yellow")
    verified = loop.run_until_complete(
        verify_gh_app_installation(api_key, base_api_url, base_ws_url, git_info)
    )

    if verified:
        click.secho("Verified!", fg="green")
    else:
        click.secho("No verification found :(", fg="red")
        sys.exit(1)

    click.secho("Creating workflow file...", fg="yellow")
    _create_workflow_yaml()


WORKFLOW_YAML = """
name: Indent Action

on:
  issue_comment:
    types: [created]
  pull_request_review_comment:
    types: [created]
  issues:
    types: [opened, assigned]
  pull_request_review:
    types: [submitted]

jobs:
  indent:
    if: |
      (github.event_name == 'issue_comment' && contains(github.event.comment.body, '@indent')) ||
      (github.event_name == 'pull_request_review_comment' && contains(github.event.comment.body, '@indent')) ||
      (github.event_name == 'pull_request_review' && contains(github.event.review.body, '@indent')) ||
      (github.event_name == 'issues' && (contains(github.event.issue.body, '@indent') || contains(github.event.issue.title, '@indent')))
    runs-on: ubuntu-latest

    steps:
      - name: Generate token for Indent app
        id: generate_token
        uses: actions/create-github-app-token@v1
        with:
          app-id: ${{ secrets.INDENT_APP_ID }}
          private-key: ${{ secrets.INDENT_APP_PRIVATE_KEY }}

      - name: Respond to mention
        uses: actions/github-script@v7
        with:
          github-token: ${{ steps.generate_token.outputs.token }}
          script: |
            const issue_number = context.payload.issue?.number || 
                               context.payload.pull_request?.number ||
                               context.payload.review?.pull_request?.number;

            await github.rest.issues.createComment({
              owner: context.repo.owner,
              repo: context.repo.repo,
              issue_number: issue_number,
              body: 'Hi it\'s me, Indent'
            });
""".lstrip()


def _create_workflow_yaml() -> None:
    git_branch = f"indent-workflow-{uuid4()}"
    workflow_file = "indent-review.yml"

    # 1. Locate the repository (searches upward until it finds .git).
    repo = Repo(Path.cwd(), search_parent_directories=True)
    if repo.bare or not repo.working_tree_dir:
        sys.exit("Error: cannot operate inside a bare repository.")

    original_branch = repo.active_branch.name if not repo.head.is_detached else None

    # 2. Create or reuse the branch, then check it out.
    try:
        branch_ref = repo.create_head(git_branch)
    except GitCommandError:
        branch_ref = repo.heads[git_branch]
    branch_ref.checkout()

    # 3. Ensure workflow directory exists.
    wf_dir = Path(repo.working_tree_dir) / ".github" / "workflows"
    wf_dir.mkdir(parents=True, exist_ok=True)

    yml = wf_dir / workflow_file
    if not yml.exists():
        yml.write_text(WORKFLOW_YAML)
        # 5. Stage & commit.
        repo.index.add([str(yml)])
        repo.index.commit("Add Indent workflow template")
        print(
            f"Created {yml.relative_to(repo.working_tree_dir)} "
            f"on branch '{git_branch}'."
        )
    else:
        print(
            f"{yml.relative_to(repo.working_tree_dir)} already exists; nothing to do."
        )

    subprocess.run(
        ["git", "push", "-u", "origin", git_branch],
        cwd=repo.working_tree_dir,
        check=True,
        capture_output=False,
        text=True,
    )

    pr_url: str | None = None
    pr_title = f"Add Indent workflow ({workflow_file})"
    pr_body = "This PR introduces an Indent github action workflow"

    def run_gh(cmd: list[str]) -> subprocess.CompletedProcess[str]:
        return subprocess.run(
            ["gh", *cmd],
            cwd=repo.working_tree_dir,
            check=True,
            capture_output=True,
            text=True,
        )

    try:
        # Does a PR already exist for this head branch?
        result = run_gh(["pr", "view", git_branch, "--json", "url"])
        pr_url = json.loads(result.stdout)["url"]
    except subprocess.CalledProcessError:
        # No PR yet → create one
        base = original_branch or repo.remotes.origin.refs[0].name.split("/")[-1]
        run_gh(
            [
                "pr",
                "create",
                "--head",
                git_branch,
                "--base",
                base,
                "--title",
                pr_title,
                "--body",
                pr_body,
            ]
        )
        # Fetch the newly created URL
        result = run_gh(["pr", "view", git_branch, "--json", "url"])
        pr_url = json.loads(result.stdout)["url"]

    if pr_url:
        click.secho(f"PR: {pr_url}", fg="green")
        webbrowser.open(pr_url)
    else:
        click.secho("Failed to create PR!", fg="red")

    if original_branch:
        repo.git.checkout(original_branch)



from collections.abc import Callable, Sequence
from gettext import gettext
from typing import Any

import click
import questionary

from exponent.core.types.generated.strategy_info import StrategyInfo


class AutoCompleteOption(click.Option):
    prompt: str

    def __init__(
        self,
        param_decls: Sequence[str] | None = None,
        prompt: bool | str = True,
        choices: list[str] | None = None,
        **kwargs: Any,
    ):
        super().__init__(param_decls, prompt=prompt, **kwargs)
        if isinstance(self.type, click.Choice):
            self.choices = self.type.choices
        else:
            self.choices = choices or []

    def prompt_for_value(self, ctx: click.core.Context) -> Any:
        return questionary.autocomplete(
            self.prompt,
            list(self.choices),
            style=questionary.Style(
                [
                    ("question", "bold"),  # question text
                    (
                        "answer",
                        "fg:#33ccff bold",
                    ),  # submitted answer text behind the question
                    (
                        "answer",
                        "bg:#000066",
                    ),  # submitted answer text behind the question
                ]
            ),
        ).unsafe_ask()


class StrategyChoice(click.Choice):
    def __init__(self, choices: Sequence[StrategyInfo]) -> None:
        self.strategy_choices = choices
        self.choices = [strategy.strategy_name.value for strategy in choices]
        self.case_sensitive = True


class StrategyOption(AutoCompleteOption):
    def __init__(self, *args: Any, type: StrategyChoice, **kwargs: Any):
        super().__init__(*args, type=type, **kwargs)
        self.default = self.default_choice(type.strategy_choices)
        self.strategy_choices = type.strategy_choices

    def _format_strategy_info(
        self, strategy_info: StrategyInfo, formatter: click.HelpFormatter
    ) -> None:
        row = (strategy_info.strategy_name.value, strategy_info.display_name)
        formatter.write_dl([row])
        with formatter.indentation():
            formatter.write_text(strategy_info.description)

    def help_extra_hook(self, formatter: click.HelpFormatter) -> None:
        with formatter.section("Strategies"):
            for strategy_info in self.strategy_choices:
                formatter.write_paragraph()
                self._format_strategy_info(strategy_info, formatter)

    @staticmethod
    def default_choice(choices: Sequence[StrategyInfo]) -> str:
        return min(choices, key=lambda x: x.display_order).strategy_name.value


class ExponentCommand(click.Command):
    def format_options(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        """Writes all the options into the formatter if they exist."""
        opts = []
        for param in self.get_params(ctx):
            rv = param.get_help_record(ctx)
            hook = getattr(param, "help_extra_hook", None)
            if rv is not None:
                opts.append((rv, hook))

        if not opts:
            return

        with formatter.section(gettext("Options")):
            for opt, hook in opts:
                formatter.write_dl([opt])
                if hook is not None:
                    hook(formatter)
                    formatter.write_paragraph()


class ExponentGroup(click.Group):
    command_class = ExponentCommand
    group_class = type


def exponent_cli_group(
    name: str | None = None,
    **attrs: Any,
) -> Callable[[Callable[..., Any]], ExponentGroup]:
    return click.command(name, ExponentGroup, **attrs)



import logging
import os
import select
import sys

from colour import Color

from exponent.utils.colors import (
    adjust_color_for_contrast,
    blend_colors_srgb,
    color_distance,
)

logger = logging.getLogger(__name__)


DEFAULT_TERM_PALETTE = [
    # Windows 10 Console default palette.
    # The least bad default of them all ;)
    # Used as a fallback when terminal doesn't report its palette,
    # which only happens on older or feature-poor terminals, i.e. very rarely.
    Color("#0c0c0c"),
    Color("#c50f1f"),
    Color("#13a10e"),
    Color("#c19c00"),
    Color("#0037da"),
    Color("#881798"),
    Color("#3a96dd"),
    Color("#cccccc"),
]


# As the only popular "modern" terminal emulator, Terminal.app doesn't support true color
# so disable true color output if we're running in it.
TRUE_COLOR = os.getenv("TERM_PROGRAM") != "Apple_Terminal"


class Theme:
    term_fg: Color
    term_bg: Color
    red: Color
    green: Color
    blue: Color
    exponent_green: Color
    hl_theme_name: str
    dimmed_text_fg: Color
    block_header_bg: Color
    block_body_bg: Color
    block_footer_fg: Color
    block_footer_bg: Color
    statusbar_default_fg: Color
    statusbar_autorun_all: Color
    statusbar_autorun_ro: Color
    statusbar_thinking_on: Color
    thinking_spinner_fg: Color

    def __init__(
        self, term_fg: Color, term_bg: Color, term_palette: list[Color | None]
    ):
        self.term_fg = term_bg
        self.term_bg = term_bg
        self.red = adjust_color_for_contrast(
            term_bg, term_palette[1] or DEFAULT_TERM_PALETTE[1]
        )
        self.green = adjust_color_for_contrast(
            term_bg, term_palette[2] or DEFAULT_TERM_PALETTE[2]
        )
        self.blue = adjust_color_for_contrast(
            term_bg, term_palette[4] or DEFAULT_TERM_PALETTE[4]
        )
        self.exponent_green = adjust_color_for_contrast(
            term_bg,
            Color("#03bd89"),  # green used in Exponent logo
        )
        dark_mode = term_bg.luminance < 0.5
        self.hl_theme_name = "ansi_dark" if dark_mode else "ansi_light"
        direction = 1 if dark_mode else -1
        (h, s, l) = term_bg.hsl  # noqa: E741
        block_target = Color(hsl=(h, s, min(1.0, l + 0.005 * direction)))
        self.block_header_bg = adjust_color_for_contrast(term_bg, block_target, 1.2)
        self.block_body_bg = adjust_color_for_contrast(term_bg, block_target, 1.1)
        self.block_footer_bg = adjust_color_for_contrast(term_bg, block_target, 1.05)
        self.block_footer_fg = blend_colors_srgb(term_fg, self.block_footer_bg, 0.4)
        self.dimmed_text_fg = adjust_color_for_contrast(
            term_bg, blend_colors_srgb(term_fg, term_bg, 0.5)
        )
        self.statusbar_default_fg = self.dimmed_text_fg
        self.statusbar_autorun_all = self.red
        self.statusbar_autorun_ro = self.green
        self.statusbar_thinking_on = self.blue
        self.thinking_spinner_fg = adjust_color_for_contrast(
            term_bg, Color("#968ce6")
        )  # nice purple


def get_term_colors(
    use_default_colors: bool,
) -> tuple[Color, Color, list[Color | None]]:
    from exponent.commands.shell_commands import POSIX_TERMINAL, RawMode

    fg = DEFAULT_TERM_PALETTE[7]
    bg = DEFAULT_TERM_PALETTE[0]
    palette: list[Color | None] = [None, None, None, None, None, None, None, None]

    # Not supported on Windows or when stdin is not a TTY
    if use_default_colors or not POSIX_TERMINAL or not sys.stdin.isatty():
        return (fg, bg, palette)

    try:
        with RawMode(sys.stdin.fileno()):
            stdin_fd = sys.stdin.fileno()
            stdout_fd = sys.stdout.fileno()

            # Try to write the ANSI escape sequences
            if stdout_fd in select.select([], [stdout_fd], [], 1)[1]:
                os.write(
                    stdout_fd,
                    b"\x1b]10;?\x07\x1b]11;?\x07\x1b]4;0;?\x07\x1b]4;1;?\x07\x1b]4;2;?\x07\x1b]4;3;?\x07\x1b]4;4;?\x07\x1b]4;5;?\x07\x1b]4;6;?\x07\x1b]4;7;?\x07\x1b[c",
                )
            else:
                return (fg, bg, palette)  # Can't write to stdout

            got_da_response = False
            timeout = 0.5  # Short timeout to avoid hanging

            # Read terminal responses until we get DA response or timeout
            while not got_da_response:
                if stdin_fd in select.select([stdin_fd], [], [], timeout)[0]:
                    reply = os.read(stdin_fd, 1024)
                    seqs = reply.split(b"\x1b")

                    for seq in seqs:
                        if seq.startswith(b"]10;rgb:"):
                            [r, g, b] = seq[8:].decode().split("/")
                            fg = Color(f"#{r[0:2]}{g[0:2]}{b[0:2]}")
                        elif seq.startswith(b"]11;rgb:"):
                            [r, g, b] = seq[8:].decode().split("/")
                            bg = Color(f"#{r[0:2]}{g[0:2]}{b[0:2]}")
                        elif seq.startswith(b"]4;"):
                            text = seq.decode()
                            idx = int(text[3])
                            [r, g, b] = text[9:].split("/")
                            palette[idx] = Color(f"#{r[0:2]}{g[0:2]}{b[0:2]}")
                        elif seq.endswith(b"c"):
                            got_da_response = True
                else:
                    # Timeout, terminal didn't respond
                    break
    except Exception:
        logger.debug("Error getting term colors", exc_info=True)
        # Any exception at all, just use default colors
        pass

    return (fg, bg, palette)


def color_component_to_8bit_cube(value: int) -> tuple[int, int]:
    # Based on https://en.wikipedia.org/wiki/ANSI_escape_code#8-bit

    if value < 48:
        value = 0
        idx = 0
    elif value < 115:
        value = 0x5F
        idx = 1
    elif value < 155:
        value = 0x87
        idx = 2
    elif value < 195:
        value = 0xAF
        idx = 3
    elif value < 235:
        value = 0xD7
        idx = 4
    else:
        value = 0xFF
        idx = 5

    return (value, idx)


def closest_8bit_color_index(c: Color) -> int:
    (quantized_red, red_idx) = color_component_to_8bit_cube(round(c.red * 255))
    (quantized_green, green_idx) = color_component_to_8bit_cube(round(c.green * 255))
    (quantized_blue, blue_idx) = color_component_to_8bit_cube(round(c.blue * 255))
    quantized = Color(
        rgb=(quantized_red / 255, quantized_green / 255, quantized_blue / 255)
    )
    quantized_error = color_distance(c, quantized)

    gray = Color(hsl=(c.hue, 0, c.luminance))
    gray_idx = round(max(gray.red * 255 - 8, 0) / 10)
    quantized_gray_value = 8 + 10 * gray_idx
    quantized_gray = Color(
        rgb=(
            quantized_gray_value / 255,
            quantized_gray_value / 255,
            quantized_gray_value / 255,
        )
    )
    quantized_gray_error = color_distance(c, quantized_gray)

    if quantized_error < quantized_gray_error:
        # 16 + 36 × r + 6 × g + b (0 ≤ r, g, b ≤ 5)
        idx = 16 + 36 * red_idx + 6 * green_idx + blue_idx
    else:
        idx = 232 + gray_idx

    return idx


def fg_color_seq(c: int | Color) -> str:
    if isinstance(c, int):
        return f"\x1b[{30 + c}m"
    elif TRUE_COLOR:
        (r, g, b) = c.rgb
        r = round(r * 255)
        g = round(g * 255)
        b = round(b * 255)
        return f"\x1b[38;2;{r};{g};{b}m"
    else:
        idx = closest_8bit_color_index(c)
        return f"\x1b[38;5;{idx}m"


def default_fg_color_seq() -> str:
    return "\x1b[39m"


def bg_color_seq(c: int | Color) -> str:
    if isinstance(c, int):
        return f"\x1b[{40 + c}m"
    elif TRUE_COLOR:
        (r, g, b) = c.rgb
        r = round(r * 255)
        g = round(g * 255)
        b = round(b * 255)
        return f"\x1b[48;2;{r};{g};{b}m"
    else:
        idx = closest_8bit_color_index(c)
        return f"\x1b[48;5;{idx}m"


def get_theme(use_default_colors: bool) -> Theme:
    (fg, bg, palette) = get_term_colors(use_default_colors)
    return Theme(fg, bg, palette)



import asyncio
import os
import shutil
from collections.abc import AsyncGenerator, Callable
from typing import Any

from exponent.core.remote_execution.languages.shell import (
    get_rc_file_source_command,
)
from exponent.core.remote_execution.languages.types import (
    ShellExecutionResult,
    StreamedOutputPiece,
)
from exponent.core.remote_execution.utils import smart_decode

STDOUT_FD = 1
STDERR_FD = 2
MAX_TIMEOUT = 300


async def read_stream(
    stream: asyncio.StreamReader, fd: int, output: list[tuple[int, str]]
) -> AsyncGenerator[StreamedOutputPiece, None]:
    data: bytes = b""
    encoding = "utf-8"

    while True:
        try:
            data = await stream.read(4096)
            if not data:
                break
            chunk = data.decode(encoding=encoding)
            output.append((fd, chunk))
            yield StreamedOutputPiece(content=chunk)
        except UnicodeDecodeError:
            decode_result = smart_decode(data)

            if decode_result is None:
                break

            # Store the detected encoding as a hint for how
            # to decode the remaining chunks of data
            chunk, encoding = decode_result

            output.append((fd, chunk))
            yield StreamedOutputPiece(content=chunk)

            continue
        except asyncio.CancelledError:
            raise
        except Exception:
            break


async def execute_shell_streaming(
    code: str,
    working_directory: str,
    timeout: int,
    should_halt: Callable[[], bool] | None = None,
) -> AsyncGenerator[StreamedOutputPiece | ShellExecutionResult, None]:
    timeout_seconds = min(timeout, MAX_TIMEOUT)

    shell_path = os.environ.get("SHELL") or shutil.which("bash") or shutil.which("sh")

    if not shell_path:
        process = await asyncio.create_subprocess_shell(
            code,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=working_directory,
        )
    else:
        # Add rc file sourcing to the command
        rc_source_cmd = get_rc_file_source_command(shell_path)
        full_command = f"{rc_source_cmd}{code}"

        process = await asyncio.create_subprocess_exec(
            shell_path,
            "-l",
            "-c",
            full_command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=working_directory,
        )

    exit_code = None
    output: list[tuple[int, str]] = []
    halted = False
    timed_out = False
    assert process.stdout
    assert process.stderr

    async def monitor_halt() -> None:
        nonlocal halted

        while True:
            if should_halt and should_halt():
                process.kill()
                halted = True
                break
            if process.returncode is not None:
                break
            await asyncio.sleep(0.1)

    def on_timeout() -> None:
        nonlocal timed_out
        timed_out = True
        process.kill()

    try:
        halt_task = asyncio.create_task(monitor_halt()) if should_halt else None
        timeout_handle = asyncio.get_running_loop().call_later(
            timeout_seconds, on_timeout
        )

        # Stream stdout and stderr concurrently using wait
        stdout_gen = read_stream(process.stdout, STDOUT_FD, output)
        stderr_gen = read_stream(process.stderr, STDERR_FD, output)

        stdout_task = asyncio.create_task(stdout_gen.__anext__())
        stderr_task = asyncio.create_task(stderr_gen.__anext__())
        pending = {stdout_task, stderr_task}

        while pending:
            done, pending = await asyncio.wait(
                pending, return_when=asyncio.FIRST_COMPLETED
            )

            for task in done:
                try:
                    piece = await task
                    yield piece

                    if process.returncode is not None:
                        break

                    # Schedule next read from the same stream
                    if task is stdout_task and not process.stdout.at_eof():
                        stdout_task = asyncio.create_task(stdout_gen.__anext__())
                        pending.add(stdout_task)
                    elif task is stderr_task and not process.stderr.at_eof():
                        stderr_task = asyncio.create_task(stderr_gen.__anext__())
                        pending.add(stderr_task)
                except StopAsyncIteration:
                    continue

        exit_code = await process.wait()
        timeout_handle.cancel()

    except asyncio.CancelledError:
        process.kill()
        raise
    finally:
        tasks_to_cancel: list[asyncio.Task[Any]] = [stdout_task, stderr_task]
        if halt_task:
            tasks_to_cancel.append(halt_task)

        await asyncio.gather(*tasks_to_cancel, return_exceptions=True)

    formatted_output = "".join([chunk for (_, chunk) in output]).strip() + "\n\n"

    yield ShellExecutionResult(
        output=formatted_output,
        cancelled_for_timeout=timed_out,
        exit_code=None if timed_out else exit_code,
        halted=halted,
    )



import os
from collections.abc import Callable
from pathlib import Path
from typing import cast

import pygit2
from anyio import Path as AsyncPath
from gitignore_parser import (
    IgnoreRule,
    handle_negation,
    parse_gitignore,
    rule_from_pattern,
)
from pygit2 import Tree
from pygit2.enums import DiffOption
from pygit2.repository import Repository

from exponent.core.remote_execution.types import (
    GitInfo,
)
from exponent.core.remote_execution.utils import safe_read_file

GIT_OBJ_COMMIT = 1


async def git_file_walk(
    repo: Repository,
    directory: str,
) -> list[str]:
    """
    Walk through a directory and return all file paths, respecting .gitignore and additional ignore patterns.
    """
    tree = get_git_subtree_for_dir(repo, directory)

    if not tree:
        return []

    # diff to the empty tree to see all files
    tracked_diff = tree.diff_to_tree()

    tracked_files = [delta.new_file.path for delta in tracked_diff.deltas]

    # Find untracked files relative to the root
    untracked_diff = repo.diff(flags=DiffOption.INCLUDE_UNTRACKED)
    untracked_files_from_root = [
        AsyncPath(delta.new_file.path) for delta in untracked_diff.deltas
    ]

    # Current working directory relative to the repo root
    dir_path = await AsyncPath(directory).resolve()
    repo_path = await AsyncPath(repo.workdir).resolve()

    if repo_path == dir_path:
        relative_directory = str(repo_path)
    else:
        relative_directory = str(dir_path.relative_to(repo_path))

    # Resolve all untracked files that are within the current working directory
    untracked_files = []
    for untracked_file in untracked_files_from_root:
        if not untracked_file.is_relative_to(relative_directory):
            continue

        untracked_files.append(str(untracked_file.relative_to(relative_directory)))

    # Combine both as sets to remove duplicates
    return list(set(tracked_files) | set(untracked_files))


def get_repo(working_directory: str) -> Repository | None:
    try:
        return Repository(working_directory)
    except pygit2.GitError:
        return None


async def get_git_info(working_directory: str) -> GitInfo | None:
    try:
        repo = Repository(working_directory)
    except pygit2.GitError:
        return None

    return GitInfo(
        branch=(await _get_git_branch(repo)) or "<unknown branch>",
        remote=_get_git_remote(repo),
    )


def get_tracked_files_in_dir(
    repo: Repository,
    dir: str | Path,
    filter_func: Callable[[str], bool] | None = None,
) -> list[str]:
    rel_path = get_path_relative_to_repo_root(repo, dir)
    dir_tree = get_git_subtree_for_dir(repo, dir)
    entries: list[str] = []
    if not dir_tree:
        return entries
    for entry in dir_tree:
        if not entry.name:
            continue
        entry_path = str(Path(f"{repo.workdir}/{rel_path}/{entry.name}"))
        if entry.type_str == "tree":
            entries.extend(get_tracked_files_in_dir(repo, entry_path, filter_func))
        elif entry.type_str == "blob":
            if not filter_func or filter_func(entry.name):
                entries.append(entry_path)
    return entries


def get_git_subtree_for_dir(repo: Repository, dir: str | Path) -> Tree | None:
    rel_path = get_path_relative_to_repo_root(repo, dir)

    try:
        head_commit = repo.head.peel(GIT_OBJ_COMMIT)
    except pygit2.GitError:
        # If the repo is empty, then the head commit will not exist
        return None
    head_tree: Tree = head_commit.tree

    if rel_path == Path("."):
        # If the relative path is the root of the repo, then
        # the head_tree is what we want. Note we do this because
        # Passing "." or "" as the path into the tree will raise.
        return head_tree
    return cast(Tree, head_tree[str(rel_path)])


def get_path_relative_to_repo_root(repo: Repository, path: str | Path) -> Path:
    path = Path(path).resolve()
    return path.relative_to(Path(repo.workdir).resolve())


def get_local_commit_hash() -> str:
    try:
        # Open the repository (assumes the current working directory is within the git repo)
        repo = Repository(os.getcwd())

        # Get the current HEAD commit
        head = repo.head

        # Get the commit object and return its hash as a string
        return str(repo[head.target].id)
    except pygit2.GitError:
        return "unknown-local-commit"


def _get_git_remote(repo: Repository) -> str | None:
    if repo.remotes:
        return str(repo.remotes[0].url)
    return None


async def _get_git_branch(repo: Repository) -> str | None:
    try:
        # Look for HEAD file in the .git directory
        head_path = AsyncPath(os.path.join(repo.path, "HEAD"))

        if not await head_path.exists():
            return None

        head_content_raw = await safe_read_file(head_path)
        head_content = head_content_raw.strip()

        if head_content.startswith("ref:"):
            return head_content.split("refs/heads/")[-1]
        else:
            return None

    except Exception:  # noqa: BLE001
        return None


class GitIgnoreHandler:
    def __init__(
        self, working_directory: str, default_ignores: list[str] | None = None
    ):
        self.checkers = {}

        if default_ignores:
            self.checkers[working_directory] = self._parse_ignore_extra(
                working_directory, default_ignores
            )

    async def read_ignorefile(self, path: str) -> None:
        new_ignore = await self._get_ignored_checker(path)

        if new_ignore:
            self.checkers[path] = new_ignore

    def filter(
        self,
        relpaths: list[str],
        root: str,
    ) -> list[str]:
        result = []

        for relpath in relpaths:
            if relpath.startswith(".git"):
                continue

            path = os.path.join(root, relpath)

            if self.is_ignored(path):
                continue

            result.append(relpath)

        return result

    def is_ignored(self, path: str) -> bool:
        return any(
            self.checkers[dp](path)
            for dp in self.checkers
            if self._is_subpath(path, dp)
        )

    def _parse_ignore_extra(
        self, working_directory: str, ignore_extra: list[str]
    ) -> Callable[[str], bool]:
        rules: list[IgnoreRule] = []

        for pattern in ignore_extra:
            if (
                rule := rule_from_pattern(pattern, base_path=working_directory)
            ) is not None:
                rules.append(rule)

        def rule_handler(file_path: str) -> bool:
            nonlocal rules
            return bool(handle_negation(file_path, rules))

        return rule_handler

    async def _get_ignored_checker(self, dir_path: str) -> Callable[[str], bool] | None:
        new_ignore = self._parse_gitignore(dir_path)

        existing_ignore = self.checkers.get(dir_path)

        if existing_ignore and new_ignore:
            return self._or(new_ignore, existing_ignore)

        return new_ignore or existing_ignore

    @staticmethod
    def _parse_gitignore(directory: str) -> Callable[[str], bool] | None:
        gitignore_path = os.path.join(directory, ".gitignore")

        if os.path.isfile(gitignore_path):
            return cast(Callable[[str], bool], parse_gitignore(gitignore_path))

        return None

    @staticmethod
    def _or(
        a: Callable[[str], bool], b: Callable[[str], bool]
    ) -> Callable[[str], bool]:
        def or_handler(file_path: str) -> bool:
            return a(file_path) or b(file_path)

        return or_handler

    @staticmethod
    def _is_subpath(path: str, parent: str) -> bool:
        """
        Check if a path is a subpath of another path.
        """
        return os.path.commonpath([path, parent]) == parent



from dataclasses import dataclass


@dataclass
class StreamedOutputPiece:
    content: str


@dataclass
class ShellExecutionResult:
    output: str
    cancelled_for_timeout: bool
    exit_code: int | None
    halted: bool = False


@dataclass
class PythonExecutionResult:
    output: str
    halted: bool = False
